from pathlib import Path

import pandas as pd


def filter_dicom(path_csv, series_description, save_path=None):
    df = pd.read_csv(path_csv, usecols=["Path", "SeriesDescription"])

    df = df.loc[df["SeriesDescription"].str.contains(series_description), "Path"]

    if save_path:
        save_path = Path(save_path)
        if not save_path.exists():
            save_path.mkdir(parents=True)

        save_path = save_path / Path("metadata.csv")

        df.to_csv(save_path, index=False)

    return df



from pathlib import Path
from typing import Iterable, List, Tuple, Union

import nibabel as nib
import numpy as np
import pydicom as dicom
import SimpleITK as sitk

PathType = Union[str, Path]
# example:
# path: PathType = Path("/storage/git/mohsen/medviz/dataset/1-1.nii")
# path: PathType = "/storage/git/mohsen/medviz/dataset/1-1.nii"
PathTypeLst = Union[List[Path], List[str]]
# example:
# paths: PathTypeLst = [Path("/storage/git/mohsen/medviz/dataset/1-1.nii"), Path("/storage/git/mohsen/medviz/dataset/1-1.nii")]
# paths: PathTypeLst = ["/storage/git/mohsen/medviz/dataset/1-1.nii", "/storage/git/mohsen/medviz/dataset/1-1.nii"]
PathTypeIter = Union[str, Path, Iterable[str], Iterable[Path]]
# example:
# paths: PathTypeIter = Path("/storage/git/mohsen/medviz/dataset/1-1.nii")
# paths: PathTypeIter = "/storage/git/mohsen/medviz/dataset/1-1.nii"
# paths: PathTypeIter = ["/storage/git/mohsen/medviz/dataset/1-1.nii", "/storage/git/mohsen/medviz/dataset/1-1.nii"]
# paths: PathTypeIter = [Path("/storage/git/mohsen/medviz/dataset/1-1.nii"), Path("/storage/git/mohsen/medviz/dataset/1-1.nii")]

StrLst = List[str]

PathNpType = Union[PathType, np.ndarray]

NumLst = List[np.ndarray]

MedicalImage = Union[nib.Nifti1Image, dicom.dataset.FileDataset, sitk.Image, np.ndarray]

PathMedicalImage = Union[PathType, MedicalImage]

TupleNp = Tuple[np.ndarray, np.ndarray]



import numpy as np
import pydicom

from ..utils import path2loader, path_in, save_path_file


#     return pydicom.tag.Tag(int(group, 16), int(element, 16))
def convert_tag(tag_str):
    if "|" in tag_str:
        group, element = tag_str.split("|")
        return pydicom.tag.Tag(int(group, 16), int(element, 16))
    else:
        return None  # Return None for strings that don't contain "|"


def get_hu(file_path, save_path=None):
    file_path = path_in(file_path)

    dicom_file = pydicom.dcmread(file_path)
    pixel_array = dicom_file.pixel_array

    rescale_slope = dicom_file.RescaleSlope
    rescale_intercept = dicom_file.RescaleIntercept

    hu_pixel_array = pixel_array * rescale_slope + rescale_intercept
    hu_pixel_array = hu_pixel_array.astype(np.int16)
    dicom_file.PixelData = hu_pixel_array.tobytes()
    if save_path is not None:
        name = f"{file_path.stem}_hu"
        save_path = save_path + "/" + name

        save_path = save_path_file(save_path, suffix=".dcm")
        dicom_file.save_as(save_path)


# def get_hu(path):
#     image = path2loader(path)
#     image_and_properties, reader, image_and_properties_itk = image

#     if reader == "pydicom":
#         ds = image_and_properties


#     else:
#         image = image_and_properties_itk

#         metadata_keys = image.GetMetaDataKeys()

#         metadata = {
#         pydicom.datadict.keyword_for_tag(convert_tag(key)): image.GetMetaData(key)
#         for key in metadata_keys
#         if convert_tag(key) is not None
#         and pydicom.datadict.keyword_for_tag(
#             convert_tag(key)
#         )  # Only include known tags with "|"
#     }


#         rescale_intercept = (
#             float(image.GetMetaData("0028|1052"))
#             if "0028|1052" in metadata_keys
#             else metadata.get("scl_inter", None)
#         )

#         rescale_slope = (
#             float(image.GetMetaData("0028|1053"))
#             if "0028|1053" in metadata_keys
#             else metadata.get("scl_slope", None)
#         )


#         metadata["itk_HU"] = rescale_intercept == -1024 and rescale_slope == 1

#         return metadata["itk_HU"]



from typing import List, Optional

import numpy as np
import SimpleITK as sitk

from ..utils import PathType, path_in, save_path_file


# change list to tuple
def resample(
    path: PathType,
    voxel_size: List[float] = [1.0, 1.0, 1.0],
    method: str = "trilinear",
    out_path: Optional[PathType] = None,
    file_name: Optional[str] = None,
):
    """
    Resamples an image to the specified voxel size using the given method.

    Parameters:
    - path (PathType): Path to the input image to be resampled.
    - voxel_size (list[float], default=[1.0, 1.0, 1.0]): Desired output voxel size.
    - method (str, default="trilinear"): Interpolation method to be used. Options include "nearest" and "trilinear".
    - out_path (Optional[PathType], default=None): Directory path to save the resampled image. If not provided, it uses the input image directory.
    - file_name (Optional[str], default=None): Name for the output resampled file. If not provided, it appends "_resampled_{method}" to the original file name.

    Returns:
    - sitk.Image: The resampled SimpleITK image.

    Raises:
    - ValueError: If an unknown interpolation method is provided.

    Notes:
    Prints out the original and resampled size and spacing information.
    """

    path = path_in(path)
    itk_image = sitk.ReadImage(path)

    original_spacing = itk_image.GetSpacing()
    original_size = itk_image.GetSize()
    print(f"original_size: {original_size}")
    print(f"original_spacing: {original_spacing}")

    out_size = [
        int(np.round(original_size[0] * (original_spacing[0] / voxel_size[0]))),
        int(np.round(original_size[1] * (original_spacing[1] / voxel_size[1]))),
        int(np.round(original_size[2] * (original_spacing[2] / voxel_size[2]))),
    ]

    resample = sitk.ResampleImageFilter()
    resample.SetOutputSpacing(voxel_size)
    resample.SetSize(out_size)
    resample.SetOutputDirection(itk_image.GetDirection())
    resample.SetOutputOrigin(itk_image.GetOrigin())
    resample.SetTransform(sitk.Transform())
    resample.SetDefaultPixelValue(itk_image.GetPixelIDValue())

    if method == "nearest":
        resample.SetInterpolator(sitk.sitkNearestNeighbor)
    elif method == "trilinear":
        resample.SetInterpolator(sitk.sitkLinear)
    else:
        raise ValueError("Unknown interpolation method")

    resampled_image = resample.Execute(itk_image)

    print(f"resampled_size: {resampled_image.GetSize()}")
    print(f"resampled_spacing: {resampled_image.GetSpacing()}")

    if not out_path:
        out_path = path.parent

    if not file_name:
        file_name = path.stem + f"_resampled_{method}" + path.suffix

    full_path = save_path_file(out_path / file_name, path.suffix)
    sitk.WriteImage(resampled_image, str(full_path))
    print(f"Saved to {full_path}")

    return resampled_image



from .custom_type import PathType, PathTypeLst, PathTypeIter, StrLst, NumLst, TupleNp
from .utility import now
from .array_profile import profile
from .array_threshold import compute_thresholds
from .helper_mask import (
    mask_characteristics,
    significant_slice_idx,
    significant_slice_idx_data,
    significant_slices,
    slice_stats,
    # mask_path_to_data_ax,
    # mask_path_to_data_sag,
    # image_path_to_data_ax,
    # image_path_to_data_sag,
    mask_expand,
    convert_mask,
    slice_stats,
    zero_out,
)

from .helper_path import path_in, save_path_file, save_path_dir

from .helper_reader import im2arr, read_image_mask, path2loader



from io import BytesIO
from pathlib import Path
from typing import List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from PIL import Image

from ..utils import path_in, save_path_file
from .plot_helpers import _generate_mask_colors, _read_images, _read_masks


def gif(
    image: Union[str, Path, np.ndarray],
    masks: Optional[
        Union[str, Path, np.ndarray, List[Union[str, Path, np.ndarray]]]
    ] = None,
    mask_colors: Optional[List[str]] = None,
    segments_title: Optional[List[str]] = None,
    start_slice: Optional[int] = 0,
    end_slice: Optional[int] = None,
    slices: Optional[List[int]] = None,
    cmap: str = "gray",
    plane: str = "axial",
    save_path: Union[str, Path] = "output.gif",
    duration: int = 100,  # Duration for each frame in the GIF (in milliseconds)
) -> None:
    """
    Creates a GIF animation from slices of a 3D image, with optional mask overlays.

    Parameters:
    - image: Path or ndarray of the 3D image.
    - masks: List of paths or ndarrays of 3D masks to overlay on the image.
    - mask_colors: Colors for each mask. Default generates random colors.
    - start_slice: Starting slice index.
    - end_slice: Ending slice index.
    - slices: Specific slice indices to use. Overrides start_slice and end_slice.
    - cmap: Colormap for displaying the image. Default is 'gray'.
    - plane: Plane of the image ('axial', 'sagittal', or 'coronal').
    - save_path: Path to save the resulting GIF.
    - duration: Duration for each frame in the GIF (in milliseconds).
    - segments_title: Titles for each mask segment.

    Returns:
    - None: The function saves the resulting GIF to save_path.
    """

    image_data = _read_images(image, plane=plane)
    if masks:
        masks_data = _read_masks(masks, plane=plane)
        mask_colors = _generate_mask_colors(len(masks_data), mask_colors)
    else:
        masks_data = [None] * len(image_data)
        mask_colors = [None] * len(image_data)

    if segments_title:
        if len(segments_title) != len(masks_data):
            print(
                "Number of segments must be equal to the number of masks, segments will be ignored"
            )
            segments_title = None

    # Validate slice indices and set defaults if not provided

    # if end_slice is None:
    #     end_slice = image_data[0].shape[0] - 1
    # end_slice = image_data[0].shape[0] - 1 if end_slice is None else end_slice

    if slices:
        slice_range = slices
    else:
        end_slice = image_data[0].shape[0] - 1 if end_slice is None else end_slice
        slice_range = range(start_slice, end_slice + 1)

    # List to store individual frames for the GIF
    frames = []

    # For each slice of the 3D image between start_slice and end_slice
    for slice_idx in slice_range:
        fig, ax = plt.subplots()

        # Display the image slice
        ax.imshow(image_data[0][slice_idx], cmap=cmap)
        title_str = ""

        # Overlay each mask on the image
        for i, mask in enumerate(masks_data):
            if mask is not None:
                ax.contour(mask[slice_idx], colors=mask_colors[i], levels=[0.5])
                if segments_title:
                    title_str += f"{segments_title[i]}: {mask_colors[i]}  "

        if segments_title:
            ax.set_title(title_str)
        ax.axis("off")

        # Convert the Matplotlib figure to a PIL Image without saving it to a file
        buf = BytesIO()
        fig.savefig(buf, format="png", bbox_inches="tight", pad_inches=0)
        buf.seek(0)
        frame = Image.open(buf)
        frames.append(frame)

        # Close the figure to free up memory
        plt.close(fig)

    # Save frames as a GIF
    gif_path = save_path_file(save_path, "gif")
    frames[0].save(
        gif_path, save_all=True, append_images=frames[1:], duration=duration, loop=0
    )



from typing import Any, Dict, List, Optional, Union

import numpy as np
import pydicom


def _metadata_nibabel(image_and_properties) -> Dict[str, Any]:
    """
    Extracts metadata from a NIfTI image using Nibabel.

    Parameters:
    image_and_properties (nib.Nifti1Image): A NIfTI image object.

    Returns:
    Dict[str, Any]: A dictionary containing image metadata.
    """
    try:
        image_info: Dict[str, Any] = {}

        image_data = image_and_properties.get_fdata()
        image_header = image_and_properties.header
        image_obj = image_and_properties.dataobj
        # affine = image_and_properties.affine

        # slope_obj, intercept_obj = image_obj.slope, getattr(image_obj, "inter", 0)

        slope_obj, intercept_obj = getattr(image_obj, "slope", None), getattr(
            image_obj, "inter", None
        )

        is_hu_obj = slope_obj == 1 and intercept_obj <= 0

        dimensions = image_data.shape
        voxel_size = image_header.get_zooms()[:3]

        # Get spatial and temporal units
        spatial_units, temporal_units = image_header.get_xyzt_units()

        # Get slope and intercept for Hounsfield Unit (HU) conversion
        slope, intercept = image_header.get_slope_inter()[:2]

        # Determine if the image is in HU format
        is_hu_format = slope == 1 and intercept <= 0

        # Check if the image is isotropic
        is_isotropic = np.allclose(voxel_size, voxel_size[0])

        # Store the information in the dictionary
        image_info = {
            "dimensions": dimensions,
            "voxel_size": voxel_size,
            "spatial_units": spatial_units,
            "temporal_units": temporal_units,
            "slope": slope,
            "slope_obj": slope_obj,
            "intercept": intercept,
            "intercept_obj": intercept_obj,
            "is_hu_obj": is_hu_obj,
            "is_hu": is_hu_format,
            "is_isotropic": is_isotropic,
        }

        return image_info
    except AttributeError as e:
        raise ValueError(f"An attribute error occurred: {e}")


def _metadata_pydicom(image_and_properties) -> Dict[str, Any]:
    """
    Extracts extended metadata from a DICOM file using pydicom.

    Parameters:
    file_path (str): The file path to the DICOM image.

    Returns:
    Dict[str, Any]: A dictionary containing extended image metadata.
    """
    try:
        # Read the DICOM file
        ds = image_and_properties

        # Extracting basic metadata
        # patient_id = getattr(ds, "PatientID", "N/A")
        # patient_name = getattr(ds, "PatientName", "N/A")
        # study_id = getattr(ds, "StudyID", "N/A")
        # study_date = getattr(ds, "StudyDate", "N/A")
        # series_number = getattr(ds, "SeriesNumber", "N/A")
        # modality = getattr(ds, "Modality", "N/A")
        # manufacturer = getattr(ds, "Manufacturer", "N/A")

        # # Extracting pixel spacing, slice thickness, image position and orientation if available
        # pixel_spacing = getattr(ds, "PixelSpacing", "N/A")
        # slice_thickness = getattr(ds, "SliceThickness", "N/A")
        # image_position = getattr(ds, "ImagePositionPatient", "N/A")
        # image_orientation = getattr(ds, "ImageOrientationPatient", "N/A")

        # # Extracting image dimensions
        # dimensions = (
        #     (int(ds.Rows), int(ds.Columns))
        #     if "Rows" in ds and "Columns" in ds
        #     else "N/A"
        # )

        # # Store the information in the dictionary
        # image_info = {
        #     "patient_id": patient_id,
        #     "patient_name": patient_name,
        #     "study_id": study_id,
        #     "study_date": study_date,
        #     "series_number": series_number,
        #     "modality": modality,
        #     "manufacturer": manufacturer,
        #     "pixel_spacing": pixel_spacing,
        #     "slice_thickness": slice_thickness,
        #     "image_position": image_position,
        #     "image_orientation": image_orientation,
        #     "dimensions": dimensions,
        # }

        image_info = {}
        # for tag in ds.keys():
        print("len(ds.keys())", len(ds.keys()))
        # exit()
        for i, tag in enumerate(ds.keys()):
            if i == 65:
                break
            element = ds.get(tag)

            if element:
                image_info[element.name] = element.value

        return image_info

    except Exception as e:
        raise RuntimeError(f"An error occurred while reading the DICOM file: {e}")


# def convert_tag(tag_str):
#     group, element = tag_str.split("|")
#     return pydicom.tag.Tag(int(group, 16), int(element, 16))
def convert_tag(tag_str):
    if "|" in tag_str:
        group, element = tag_str.split("|")
        return pydicom.tag.Tag(int(group, 16), int(element, 16))
    else:
        return None  # Return None for strings that don't contain "|"


def _metadata_sitk(image_and_properties) -> Dict[str, Any]:
    """
    Extracts metadata from an image file using SimpleITK.

    Parameters:
    file_path (str): The file path to the image.

    Returns:
    Dict[str, Any]: A dictionary containing image metadata.
    """
    try:
        # Read the image file
        # image = sitk.ReadImage(file_path)
        image = image_and_properties

        # Extracting metadata
        metadata_keys = image.GetMetaDataKeys()

        # metadata = {key: image.GetMetaData(key) for key in metadata_keys}
        # metadata = {
        #     pydicom.datadict.keyword_for_tag(key): image.GetMetaData(key)
        #     for key in metadata_keys
        # }

        # metadata = {
        #     pydicom.datadict.keyword_for_tag(convert_tag(key)): image.GetMetaData(key)
        #     for key in metadata_keys
        #     if pydicom.datadict.keyword_for_tag(
        #         convert_tag(key)
        #     )  # Only include known tags
        # }

        metadata = {
            pydicom.datadict.keyword_for_tag(convert_tag(key)): image.GetMetaData(key)
            for key in metadata_keys
            if convert_tag(key) is not None
            and pydicom.datadict.keyword_for_tag(
                convert_tag(key)
            )  # Only include known tags with "|"
        }

        # exit()
        rescale_intercept = (
            float(image.GetMetaData("0028|1052"))
            if "0028|1052" in metadata_keys
            else metadata.get("scl_inter", None)
        )

        rescale_slope = (
            float(image.GetMetaData("0028|1053"))
            if "0028|1053" in metadata_keys
            else metadata.get("scl_slope", None)
        )

        metadata["itk_HU"] = rescale_intercept == -1024 and rescale_slope == 1

        metadata["itk_rescale_intercept"] = rescale_intercept
        metadata["itk_rescale_slope"] = rescale_slope
        # Adding additional information about the image
        metadata["itk_spacing"] = image.GetSpacing()
        metadata["itk_origin"] = image.GetOrigin()
        metadata["itk_direction"] = image.GetDirection()
        metadata["itk_size"] = image.GetSize()
        metadata[
            "itk_number_of_components_per_pixel"
        ] = image.GetNumberOfComponentsPerPixel()

        return metadata
    except RuntimeError as e:
        raise RuntimeError(f"An error occurred while reading the image file: {e}")



from pathlib import Path

import pandas as pd

from ..utils import get_characteristics, path_in, save_path_file, significant_slice_idx


def generate_mask_schema(
    masks_path: str or Path,
    id_func=lambda x: x,
    mask_extension=None,
    save_path: str or Path = Path("./output/mask_profile.csv"),
):
    df_dict = {
        "id": [],
        "path": [],
        "shape": [],
        "data_type": [],
        "values": [],
        "mask_type": [],
        "most_value_nonzero_slices": [],
        "num_nonzero_slices": [],
    }

    dataset_path = path_in(masks_path, env=False)

    if mask_extension is None:
        extensions = set()

        for file_path in dataset_path.rglob("*"):
            if file_path.is_file():
                file_extension = file_path.suffix.lower()

                extensions.add(file_extension)

        for extension in extensions:
            if extension in [".nii", ".nii.gz", ".dcm"]:
                mask_extension = extension
                break

    mask_paths = dataset_path.glob(rf"**/*{mask_extension}")

    for mask_path in mask_paths:
        # try:
        id = id_func(mask_path.stem)
        print(id)
        mask_characteristics = get_characteristics(mask_path)
        most_value_nonzero_slices, num_nonzero_slices = significant_slice_idx(mask_path)
        df_dict["id"].append(id)
        df_dict["path"].append(Path(*mask_path.parts[-3:]))
        df_dict["shape"].append(mask_characteristics["shape"])
        df_dict["data_type"].append(mask_characteristics["data_type"])
        df_dict["values"].append(mask_characteristics["values"])
        df_dict["mask_type"].append(mask_characteristics["mask_type"])
        df_dict["most_value_nonzero_slices"].append(most_value_nonzero_slices)
        df_dict["num_nonzero_slices"].append(num_nonzero_slices)

    # except Exception as e:
    #     print(f"Error in {mask_path}", e)

    df = pd.DataFrame(df_dict, columns=df_dict.keys())

    save_path = save_path_file(save_path, suffix=".csv")

    df.to_csv(save_path, index=False)

    print("Done!")



import math
from pathlib import Path
from typing import List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.widgets import Slider

from ..utils import save_path_file
from .plot_helpers import _generate_mask_colors, _read_images, _read_masks

# mask_characteristics = mask_characteristics(mask_path)


# def _generate_mask_colors(num_masks, mask_colors=None):
#     if mask_colors is None:
#         mask_colors = ["#" + secrets.token_hex(3) for _ in range(num_masks)]

#         return mask_colors
#     else:
#         try:
#             assert num_masks == len(mask_colors)
#             return mask_colors
#         except AssertionError:
#             raise ValueError(
#                 f"Number of masks ({num_masks}) does not match number of colors ({len(mask_colors)})"
#             )


# def _read_images(images, plane="axial"):
#     images_data = []
#     if not isinstance(images, (list, tuple)):
#         images = [images]
#     for image in images:
#         if isinstance(image, Union[str, Path]):
#             image = path_in(image)
#             image = sitk.ReadImage(str(image))
#             image = sitk.GetArrayFromImage(image)

#             if plane == "sagittal":
#                 image = image.transpose(1, 2, 0)
#             elif plane == "coronal":
#                 image = image.transpose(2, 1, 0)

#             images_data.append(image)
#         elif isinstance(image, np.ndarray):
#             images_data.append(image)
#         else:
#             raise ValueError("images must be a list of strings or numpy arrays")
#     return images_data


# def _read_masks(masks, plane="axial"):
#     masks_data = []
#     if masks is not None and not isinstance(masks, (list, tuple)):
#         masks = [masks]

#     for mask in masks:
#         # if mask is not None and mask_characteristics(mask)["mask_type"] != "Binary":
#         #     print(f"Mask {mask} is not binary.")

#         if isinstance(mask, Union[str, Path]):
#             mask = path_in(mask)
#             mask = sitk.ReadImage(str(mask))
#             mask = sitk.GetArrayFromImage(mask)
#             if plane == "sagittal":
#                 mask = mask.transpose(1, 2, 0)
#             elif plane == "coronal":
#                 mask = mask.transpose(2, 1, 0)

#             masks_data.append(mask)
#         elif isinstance(mask, np.ndarray):
#             masks_data.append(mask)
#         elif not mask:
#             masks_data.append(mask)
#         else:
#             raise ValueError("masks must be a list of strings or numpy arrays")
#     return masks_data


def plot3d(
    images: Union[str, Path, np.ndarray, List[Union[str, Path, np.ndarray]]],
    masks: Optional[
        Union[
            str,
            Path,
            np.ndarray,
            List[Union[str, Path, np.ndarray]],
        ]
    ] = None,
    rows: Optional[int] = None,
    columns: Optional[int] = None,
    mask_colors: Optional[List[str]] = None,
    cmap: str = "gray",
    titles: Optional[Union[str, List[str]]] = None,
    plane: str = "axial",
    save_path: Optional[Union[str, Path]] = None,
) -> None:
    images = _read_images(images, plane=plane)
    num_images = len(images)
    assert num_images > 0, "images cannot be empty"

    if masks:
        masks = _read_masks(masks, plane=plane)

        assert num_images == len(masks), "Number of images and masks must be equal"

        mask_colors = _generate_mask_colors(num_images, mask_colors)
    else:
        masks = [None] * num_images
        mask_colors = [None] * num_images

    if titles:
        assert num_images == len(titles), "Number of images and titles must be equal"
    else:
        titles = [f"Image {i}" for i in range(num_images)]

    if rows is None and columns is None:
        rows = math.ceil(math.sqrt(num_images))  # Number of rows in the grid
        columns = math.ceil(num_images / rows)  # Number of columns in the grid
    elif rows is None:
        rows = math.ceil(num_images / columns)
    elif columns is None:
        columns = math.ceil(num_images / rows)

    if num_images > rows * columns:
        raise ValueError("rows * columns must be greater than or equal to num_images")

    _, axs = plt.subplots(rows, columns)

    if num_images == 1:
        axs = np.array([axs])

    plt.subplots_adjust(bottom=0.25)

    sliders = []
    for i, ax in enumerate(axs.flat):
        if i < num_images:
            middle_slice = images[i].shape[0] // 2
            depth = images[i].shape[0]
            ax.imshow(images[i][middle_slice], cmap=cmap)
            if masks and masks[i] is not None:
                ax.contour(masks[i][middle_slice], colors=mask_colors[i], levels=[0.5])

            # title = titles[i] if titles else f"Image {i}"
            ax.set_title(titles[i])
            ax.set_xlabel(f"Slice Number: {middle_slice}")

            # slider_ax = plt.axes([0.1, 0.05 * (i + 1), 0.8, 0.03])
            # slider_position = ax.get_position()
            # slider_ax = plt.axes([slider_position.x0, slider_position.y0-0.1, slider_position.width, 0.03])

            slider_position = ax.get_position()
            slider_ax = plt.axes(
                [
                    slider_position.x0,
                    slider_position.y0 - 0.15,
                    slider_position.width,
                    0.03,
                ]
            )

            slider = Slider(
                slider_ax,
                f"Slice {i+1}",
                0,
                depth - 1,
                valinit=middle_slice,
                valstep=1,
            )
            sliders.append(slider)

            # print(slider_ax)

            def update(
                val,
                ax=ax,
                img=images[i],
                mask=masks[i],
                title=titles[i],
                # mask = None,
                slider=slider,
                color=mask_colors[i],
                # color = None,
            ):
                # if masks is not None:
                #     mask = masks[i]
                #     color = mask_colors[i]

                slice_num = int(slider.val)
                ax.clear()
                ax.imshow(img[slice_num], cmap=cmap)
                if mask is not None:
                    ax.contour(mask[slice_num], colors=color, levels=[0.5])

                ax.set_title(title)
                ax.set_xlabel(f"Slice Number: {slice_num}")

            slider.on_changed(update)
        else:
            ax.axis("off")

    plt.tight_layout()
    plt.show()

    if save_path:
        save_path = save_path_file(save_path, suffix=".png")
        plt.savefig(save_path)
    else:
        plt.show()

    plt.close()



import numpy as np
from scipy.stats import kurtosis, skew

from .main import Collage


def compute_collage(
    image, mask, haralick_windows=[3, 5, 7, 9, 11], raw_save_path=None, name=None
):
    """This method computes Collage features

    Args:
        image (path object): path of image
        mask (path object): path of mask
        haralick_windows (list, optional): Kernel window size. Defaults to [3, 5, 7, 9, 11].

    Returns:
        numpy array: mean, standard deviation, skewness, and kurtosis of COLLAGE features based on kernel size and orientation
    """
    windows_length = len(haralick_windows)

    feats = np.zeros((13 * windows_length * 2, 4), dtype=np.double)

    for window_idx, haralick_window_size in enumerate(haralick_windows):
        try:
            collage = Collage(
                image,
                mask,
                svd_radius=5,
                verbose_logging=True,
                num_unique_angles=64,
                haralick_window_size=haralick_window_size,
            )

            collage_feats = collage.execute()
            if raw_save_path:
                raw_file_path = (
                    raw_save_path / f"{name}_COLLAGE_RAW_W{haralick_window_size}.npy"
                )

                np.save(
                    raw_file_path,
                    collage_feats,
                )
                print("saved")

            for orientation in range(2):
                for collage_idx in range(13):
                    k = window_idx * collage_idx * orientation
                    feat = collage_feats[:, :, :, collage_idx, orientation].flatten()
                    feat = feat[~np.isnan(feat)]

                    feats[k, 0] = feat.mean()
                    feats[k, 1] = feat.std()
                    feats[k, 2] = skew(feat)
                    feats[k, 3] = kurtosis(feat)

        except ValueError as err:
            print(f"VALUE ERROR- {err}")

        except Exception as err:
            print(f"EXCEPTION- {err}")

    return feats



from pathlib import Path

import pandas as pd
import pydicom


def read_dicom_metadata(filepath):
    ds = pydicom.dcmread(filepath)
    metadata = {}

    for element in ds.iterall():
        if (
            element.VR != "SQ"
            and element.VR != "OB"
            and element.VR != "OW"
            and element.VR != "OF"
            and element.VR != "UN"
        ):
            metadata[element.keyword] = element.value

    return metadata


def save_dicom_metadata(
    base_path, id_func=lambda x: x, pattern="**/*.dcm", save_path: str or Path = "./"
):
    paths = Path(base_path).glob(pattern)
    metadata_list = []

    for path in paths:
        id = id_func(path.stem)

        metadata = read_dicom_metadata(path)
        metadata["ID"] = id
        metadata["Path"] = str(path)
        metadata_list.append(metadata)

    df = pd.DataFrame(metadata_list)

    save_path = Path(save_path)
    if not save_path.exists():
        save_path.mkdir(parents=True)

    save_path = save_path / Path("metadata.csv")

    df.to_csv(save_path, index=False)



from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from tabulate import tabulate

from ..utils import path2loader, path_in, save_path_file
from .meta_helpers import _metadata_nibabel, _metadata_pydicom, _metadata_sitk


def metadata(
    image_paths: Union[List[Union[str, Path]], Union[str, Path]],
    save_csv: Optional[Union[str, Path]] = None,
) -> List[Dict[str, Any]]:
    """
    Extracts and optionally saves metadata from one or more NIFTI images.

    Parameters:
        image_paths: Union[List[Union[str, Path]], Union[str, Path]] - The paths to the NIFTI images.
        save_csv: Optional[Union[str, Path]] - The path where to save the metadata as a CSV file.

    Returns:
        A list of dictionaries, each containing the metadata for one image.
    """

    if not isinstance(image_paths, list):
        image_paths = [image_paths]

    all_image_info = []

    for image_path in image_paths:
        image_path = path_in(image_path)
        # image_info: Dict[str, Any] = {}

        # try:
        #     image = nib.load(image_path)
        # except nib.filebasedimages.ImageFileError:
        #     print(f"Error: {image_path} is not a valid NIFTI image.")
        #     continue

        try:
            image_and_properties, reader, image_and_properties_itk = path2loader(
                image_path
            )

        except Exception as e:
            print("Error reading image", e)

            raise e

        merged_dict = {
            "name": image_path.name,
            **_metadata_sitk(image_and_properties_itk),
        }

        if reader == "nibabel":
            metadata_nib = _metadata_nibabel(image_and_properties)
            merged_dict = {**merged_dict, **metadata_nib}

        elif reader == "pydicom":
            metadata_dicom = _metadata_pydicom(image_and_properties)
            merged_dict = {**merged_dict, **metadata_dicom}

        all_image_info.append(merged_dict)

    if save_csv:
        csv_path = save_path_file(save_csv, suffix=".csv")
        df = pd.DataFrame(all_image_info)
        df.to_csv(csv_path, index=False)

    else:
        for info in all_image_info:
            print(tabulate(info.items(), tablefmt="fancy_grid"))

    return all_image_info



import ast
from pathlib import Path

import nibabel as nib
import numpy as np
import pandas as pd

from medviz.utils.helper_path import save_path_dir

from .compute_collage import compute_collage


def compute_stats_collage(
    csv_path: Path or str,
    stats_save_path: Path or str,
    base_path=None,
    raw_save_path: Path or str = None,
):
    stats_save_path = save_path_dir(stats_save_path)
    raw_save_path = save_path_dir(raw_save_path) if raw_save_path else None

    df = pd.read_csv(csv_path)
    for idx, row in df.iterrows():
        print(f"Processing {row['ID']} ...")

        path_image = ast.literal_eval(row["Image"])
        path_masks = ast.literal_eval(row["Mask"])

        path_image = Path(path_image[0])
        path_image = path_image if base_path is None else base_path / path_image

        for path_mask in path_masks:
            try:
                path_mask = Path(path_mask)
                path_mask = path_mask if base_path is None else base_path / path_mask
                mask_name = Path(path_mask).stem

                extension = path_mask.suffix

                if extension == ".npy":
                    mask = np.load(path_mask)
                    image = np.load(path_image)
                elif extension in [".nii", ".nii.gz"]:
                    image = nib.load(path_image).get_fdata()
                    mask = nib.load(path_mask).get_fdata()
                else:
                    raise ValueError("Unknown extension")

                image = np.swapaxes(image, 0, 2)
                mask = np.swapaxes(mask, 0, 2)

                feats = compute_collage(
                    image,
                    mask,
                    haralick_windows=[3, 5, 7, 9, 11],
                    raw_save_path=raw_save_path,
                    name=mask_name,
                )

                stats_save_file = stats_save_path / f"{idx}_{mask_name}.npy"

                np.save(stats_save_file, feats)

            except FileNotFoundError as e:
                print("File not found:", e)

            except Exception as e:
                print("Exception occurred:", e)



import secrets
from pathlib import Path
from typing import Union

import numpy as np
import SimpleITK as sitk

from ..utils import path_in


def _generate_mask_colors(num_masks, mask_colors=None):
    if mask_colors is None:
        mask_colors = ["#" + secrets.token_hex(3) for _ in range(num_masks)]

        return mask_colors
    else:
        try:
            assert num_masks == len(mask_colors)
            return mask_colors
        except AssertionError:
            raise ValueError(
                f"Number of masks ({num_masks}) does not match number of colors ({len(mask_colors)})"
            )


def _read_images(images, plane="axial"):
    images_data = []
    if not isinstance(images, (list, tuple)):
        images = [images]
    for image in images:
        if isinstance(image, Union[str, Path]):
            # image = path_in(image)
            image = sitk.ReadImage(str(image))
            image = sitk.GetArrayFromImage(image)

            if plane == "sagittal":
                image = image.transpose(1, 2, 0)
            elif plane == "coronal":
                image = image.transpose(2, 1, 0)

            images_data.append(image)
        elif isinstance(image, np.ndarray):
            images_data.append(image)
        else:
            raise ValueError("images must be a list of strings or numpy arrays")
    return images_data


def _read_masks(masks, plane="axial"):
    masks_data = []
    if masks is not None and not isinstance(masks, (list, tuple)):
        masks = [masks]

    for mask in masks:
        # if mask is not None and mask_characteristics(mask)["mask_type"] != "Binary":
        #     print(f"Mask {mask} is not binary.")

        if isinstance(mask, Union[str, Path]):
            mask = path_in(mask)
            mask = sitk.ReadImage(str(mask))
            mask = sitk.GetArrayFromImage(mask)
            if plane == "sagittal":
                mask = mask.transpose(1, 2, 0)
            elif plane == "coronal":
                mask = mask.transpose(2, 1, 0)

            masks_data.append(mask)
        elif isinstance(mask, np.ndarray):
            masks_data.append(mask)
        elif not mask:
            masks_data.append(mask)
        else:
            raise ValueError("masks must be a list of strings or numpy arrays")
    return masks_data




def reader(path: PathMedicalImage) -> PathMedicalImage:
    """
    Read the medical image based on its file extension.

    This function supports various medical image file formats, including
    but not limited to NIfTI (`.nii`, `.nii.gz`), DICOM (`.dcm`), MHA (`.mha`),
    and NumPy arrays (`.npy`, `.npz`).

    Parameters:
        path (PathMedicalImage): Path to the medical image file or an already loaded image object.
                                 Expected types include str, Path, or specific medical image objects
                                 like MedicalImage.

    Returns:
        object: Loaded medical image object.

    Raises:
        TypeError: If the provided file format is unsupported.
    """
    if isinstance(path, MedicalImage):
        return path

    path = path_in(path)
    if path.suffix in [".nii", ".nii.gz"]:
        return nib.load(path)
    elif path.suffix == ".dcm":
        return dicom.dcmread(path)
    elif path.suffix == ".mha":
        return sitk.ReadImage(str(path))
    elif path.suffix == ".npy":
        return np.load(path)
    elif path.suffix == ".npz":
        return np.load(path)["arr_0"]
    else:
        raise TypeError(f"Unsupported file type: {path.suffix}")


def reader(*paths: Union[str, Path]) -> Union[object, Tuple[object]]:
    """
    Read one or more medical images based on their file extensions.

    This function supports various medical image file formats, including
    but not limited to NIfTI (`.nii`, `.nii.gz`), DICOM (`.dcm`), MHA (`.mha`),
    and NumPy arrays (`.npy`, `.npz`).

    Parameters:
        *paths (Union[str, Path]): Paths to the medical image files or already loaded image objects.

    Returns:
        Union[object, Tuple[object]]: If a single path is provided, returns a single loaded medical image object.
                                      If multiple paths are provided, returns a tuple of loaded medical image objects.

    Raises:
        TypeError: If the provided file format is unsupported.
    """

    def load_image(path: Union[str, Path]) -> object:
        if isinstance(path, MedicalImage):
            return path

        path_obj = path_in(path)

        if path_obj.suffix in [".nii", ".nii.gz"]:
            return nib.load(path_obj)
        elif path_obj.suffix == ".dcm":
            return dicom.dcmread(path_obj)
        elif path_obj.suffix == ".mha":
            return sitk.ReadImage(str(path_obj))
        elif path_obj.suffix == ".npy":
            return np.load(path_obj)
        elif path_obj.suffix == ".npz":
            return np.load(path_obj)["arr_0"]
        else:
            raise TypeError(f"Unsupported file type: {path_obj.suffix}")

    images = tuple(load_image(path) for path in paths)
    if len(images) == 1:
        return images[0]
    return images


from os import getenv
from pathlib import Path
from typing import Optional

from .custom_type import PathType
from .utility import now


def path_in(path: PathType, env: bool = False) -> Path:
    """
    Converts a given path string or Path object into an absolute Path object.
    If env is set to True, it uses the ROOT_IN environment variable as a base directory.

    Parameters:
    - path (Union[str, Path]): The input path as a string or Path object.
    - env (bool, optional): Flag to indicate if the ROOT_IN environment variable
      should be used as a base directory. Defaults to False.

    Returns:
    - Path: The absolute path.

    Raises:
    - TypeError: If the input is neither a string nor a Path object.
    - FileNotFoundError: If the resulting path doesn't exist.
    """

    if isinstance(path, str):
        path = Path(path)

    if not isinstance(path, Path):
        raise TypeError(f"The input must be string or Path, not {type(path)}")

    if not path.is_absolute() and env:
        root_in = getenv("ROOT_IN") or "."
        path = Path(root_in) / path

    if not path.exists():
        raise FileNotFoundError(f"{path} doesn't exist on the earth.")

    return path


def save_path_dir(path: PathType, env: bool = False) -> Path:
    """
    Ensures the provided path string or Path object for an output directory exists.
    If the path does not exist, it creates the directory. If env is set to True,
    the ROOT_OUT environment variable is used as a base directory.

    Parameters:
    - path (Union[str, Path]): The output directory path as a string or Path object.
    - env (bool, optional): Flag to indicate if the ROOT_OUT environment variable
      should be used as a base directory. Defaults to False.

    Returns:
    - Path: The absolute path to the output directory.

    Raises:
    - TypeError: If the input is neither a string nor a Path object.
    """

    if isinstance(path, str):
        path = Path(path)

    if not isinstance(path, Path):
        raise TypeError(f"The output path must be string or Path, not {type(path)}")

    if not path.is_absolute() and env:
        root_out = getenv("ROOT_OUT") or "."
        path = Path(root_out) / path

    if not path.exists():
        print("The output path doesn't exist but no worries I'm making it.")
        path.mkdir(parents=True)

    return path


# def path_out_file(path: PathType, suffix: str or None, env: bool = False) -> Path:
#     if isinstance(path, str):
#         path = Path(path)

#     if not isinstance(path, Path):
#         raise TypeError(f"The output path must be string or Path, not {type(path)}")

#     if not path.is_absolute() and env:
#         root_out = getenv("ROOT_OUT") or "."
#         path = Path(root_out) / path

#     if not path.suffix:
#         if not suffix:
#             raise ValueError("suffix must be provided if save_path has no suffix")
#         path = path.with_suffix(suffix)

#     if not path.parent.exists():
#         print("The output path doesn't exist but no worries I'm making it.")
#         path.parent.mkdir(parents=True)

#     return path


def save_path_file(
    save_path: PathType, suffix: Optional[str] = None, env: Optional[bool] = False
) -> Path:
    """
    Generates a save path for a file, ensuring the directory exists and providing
    a suffix if necessary. If the file already exists, a timestamp is appended to
    avoid overwriting.

    Parameters:
    - save_path (PathType): The initial path (either as a string or Path object)
      where the file should be saved.
    - suffix (Optional[str], default=None): A file extension to be added if
      `save_path` lacks one. If `save_path` doesn't have a suffix and this
      parameter is not provided, a ValueError is raised.
    - env (Optional[bool], default=False): If set to True and `save_path` is not
      absolute, the function uses the ROOT_OUT environment variable as a base
      directory to construct the absolute path.

    Returns:
    - Path: The absolute path where the file should be saved.

    Raises:
    - TypeError: If `save_path` is neither a string nor a Path object.
    - ValueError: If `save_path` lacks a suffix and no suffix is provided.

    Note:
    If the computed `save_path` directory does not exist, it will be created.
    If a file already exists at the computed path, a timestamp is appended to
    the filename to ensure uniqueness.
    """

    if isinstance(save_path, str):
        save_path = Path(save_path)

    if not isinstance(save_path, Path):
        raise TypeError(f"save_path must be a str or Path, not {type(save_path)}")

    if not save_path.is_absolute() and env:
        root_out = getenv("ROOT_OUT") or "."
        save_path = Path(root_out) / save_path

    if not save_path.suffix:
        if not suffix:
            raise ValueError("suffix must be provided if save_path has no suffix")
        save_path = save_path.with_suffix(suffix)

    if not save_path.parent.exists():
        print("The output path doesn't exist but no worries I'm making it.")
        save_path.parent.mkdir(parents=True)
    elif save_path.exists():
        save_path = save_path.with_stem(save_path.stem + "_" + now())

    return save_path



import numpy as np
from skimage.filters import (
    threshold_li,
    threshold_mean,
    threshold_minimum,
    threshold_otsu,
    threshold_triangle,
    threshold_yen,
)


def compute_thresholds(image):
    thresholds = {}

    thresholds["otsu"] = threshold_otsu(image)
    thresholds["li"] = threshold_li(image)
    thresholds["triangle"] = threshold_triangle(image)
    thresholds["yen"] = threshold_yen(image)
    thresholds["mean"] = threshold_mean(image)
    thresholds["minimum"] = threshold_minimum(image)
    thresholds["median"] = np.median(image)
    thresholds["min"] = np.min(image)
    thresholds["max"] = np.max(image)

    return thresholds



import os

import nibabel as nib
import numpy as np
import pydicom


def _create_affine(slices):
    spacing = map(float, ([slices[0].SliceThickness] + slices[0].PixelSpacing))
    spacing = np.array(list(spacing))
    affine = np.diag(np.append(spacing, [1]))

    return affine


def series2whole(dicom_dir, save_path):
    try:
        slices = [
            pydicom.dcmread(os.path.join(dicom_dir, f)) for f in os.listdir(dicom_dir)
        ]
        slices.sort(key=lambda x: float(x.ImagePositionPatient[2]))

        images = np.stack([s.pixel_array for s in slices])

        transposed_images = np.transpose(images, (1, 2, 0))

        nib.save(nib.Nifti1Image(transposed_images, np.eye(4)), save_path)

    except Exception as e:
        print(e)




def mohsen1(
    path: PathType,
    method: str,
    voxel_size=[1, 1, 1],
    save=True,
    save_path=None,
    out_name: str = None,
) -> sitk.Image:
    path = path_in(path)

    if method == "trilinear":
        interpolator = sitk.sitkLinear
    elif method == "nearest":
        interpolator = sitk.sitkNearestNeighbor
    else:
        raise ValueError("Unknown interpolation method")

    image = sitk.ReadImage(path)

    resampler = sitk.ResampleImageFilter()
    resampler.SetOutputSpacing(voxel_size)

    print(f"Original spacing: {image.GetSpacing()}")
    print(f"Original size: {image.GetSize()}")

    # resampler.SetSize(
    #     [
    #         int(sz / vx * voxel_size[1])
    #         for sz, vx in zip(image.GetSize(), image.GetSpacing())
    #     ]
    # )

    # resampler.SetOutputDirection(image.GetDirection())
    # resampler.SetOutputOrigin(image.GetOrigin())
    # resampler.SetInterpolator(interpolator)

    # resampler.SetSize(
    #     [
    #         int(sz / vx * nv)
    #         for sz, vx, nv in zip(image.GetSize(), image.GetSpacing(), voxel_size)
    #     ]
    # )

    resampler = sitk.ResampleImageFilter()
    resampler.SetOutputSpacing(voxel_size)
    resampler.SetOutputDirection(image.GetDirection())
    resampler.SetOutputOrigin(image.GetOrigin())
    resampler.SetInterpolator(interpolator)
    resampler.SetSize(
        [
            int(sz / vx * nv)
            for sz, vx, nv in zip(image.GetSize(), image.GetSpacing(), voxel_size)
        ]
    )

    resampled_image = resampler.Execute(image)

    print(f"New spacing: {resampled_image.GetSpacing()}")
    print(f"New size: {resampled_image.GetSize()}")


# def mohsen(
#     path: PathType,
#     method: str,
#     voxel_size=[1, 1, 1],
#     save=True,
#     save_path=None,
#     out_name: str = None,
# ) -> sitk.Image:
#     path = path_in(path)

#     if method == "trilinear":
#         interpolator = sitk.sitkLinear
#     elif method == "nearest":
#         interpolator = sitk.sitkNearestNeighbor
#     else:
#         raise ValueError("Unknown interpolation method")

#     image = sitk.ReadImage(path)

#     # Explicitly calculate new sizes
#     new_sizes = [
#         int(old_size * (old_spacing / new_spacing))
#         for old_size, old_spacing, new_spacing in zip(image.GetSize(), image.GetSpacing(), voxel_size)
#     ]

#     resampler = sitk.ResampleImageFilter()
#     resampler.SetOutputSpacing(voxel_size)
#     resampler.SetOutputDirection(image.GetDirection())
#     resampler.SetOutputOrigin(image.GetOrigin())
#     resampler.SetInterpolator(interpolator)
#     resampler.SetSize(new_sizes)  # Set sizes explicitly

#     resampled_image = resampler.Execute(image)

#     print(f"Original spacing: {image.GetSpacing()}")
#     print(f"Original size: {image.GetSize()}")
#     print(f"New spacing: {resampled_image.GetSpacing()}")
#     print(f"New size: {resampled_image.GetSize()}")

#     if save:
#         suffix = path.suffix

#         if not save_path:
#             save_path = path.parent

#         if not out_name:
#             stem = path.stem
#             out_name = f"{stem}_isotropic{suffix}"

#             out_path = Path(save_path) / out_name

#             full_path = save_path_file(out_path, suffix)

#             sitk.WriteImage(resampled_image, str(full_path))
#             print(f"Saved to {full_path}")

#     return resampled_image



# from .utils.helper_mask import *

# # from .preprocess.mask import generate_mask_schema

# from .preprocess.mask import generate_mask_schema
# from .preprocess.match_image_mask import match_image_mask

from .utils import (
    # image_path_to_data_ax,
    # image_path_to_data_sag,
    # mask_path_to_data_ax,
    # mask_path_to_data_sag,
    save_path_dir,
    save_path_file,
    path_in,
    im2arr,
    read_image_mask,
    path2loader,
    # mask_characteristics,
)

from .plot import plot3d, gif

from .utils.array_profile import profile
from .utils.array_threshold import compute_thresholds

# from .utils import significant_slices

from .multimodal import save_dicom_metadata, filter_dicom

from .preprocess import *

from .feats import *

# from .nifti.helper import nii3d_to_annotated2d, nii_mask3d_to_2d

# def version():
#     with open("VERSION", "r") as f:
#         version = f.read().strip()
#     return version

# __version__ = version()

__version__ = "1.1.8"



import logging
import pickle
from pathlib import Path
from typing import List, Union

import numpy as np
from scipy.stats import kurtosis, skew

from ...utils import significant_slices
from .main import Collage

logger = logging.getLogger()

logger.info("Collage feature extraction")

descriptors = [
    "AngularSecondMoment",  # 0
    "Contrast",  # 1
    "Correlation",  # 2
    "SumOfSquareVariance",  # 3
    "SumAverage",  # 4
    "SumVariance",  # 5
    "SumEntropy",  # 6
    "Entropy",  # 7
    "DifferenceVariance",  # 8
    "DifferenceEntropy",  # 9
    "InformationMeasureOfCorrelation1",  # 10
    "InformationMeasureOfCorrelation2",  # 11
    "MaximalCorrelationCoefficient",  # 12
]


def compute_collage3d(
    image: np.ndarray, mask: np.ndarray, haralick_windows: List[int]
) -> np.ndarray:
    feats = {}

    try:
        collage = Collage(
            image,
            mask,
            svd_radius=5,
            verbose_logging=True,
            num_unique_angles=64,
            haralick_window_size=haralick_windows,
        )

        collage_feats = collage.execute()

        print("Collage feats shape", collage_feats.shape)

        for collage_idx, descriptor in enumerate(descriptors):
            print(f"Processing collage {descriptor}")
            for orientation in range(2):
                feat = collage_feats[:, :, :, collage_idx, orientation].flatten()
                feat = feat[~np.isnan(feat)]

                feats[f"col_des_{descriptor}_ori_{orientation}"] = [
                    feat.mean(),
                    feat.std(),
                    skew(feat),
                    kurtosis(feat),
                ]

    except ValueError as err:
        print(f"VALUE ERROR- {err}")

    except Exception as err:
        print(f"EXCEPTION- {err}")

    return feats


def collage3d(
    image: np.ndarray,
    mask: np.ndarray,
    window_sizes: List[int],
    save_path,
    out_name: str,
    padding: Union[bool, int] = False,
) -> None:
    """
    Process a 3D image and its associated mask using a collage method with
    specified window sizes, and save the resulting features.

    Parameters:
    -----------
    image : np.ndarray
        The 3D image data to be processed.

    mask : np.ndarray
        The 3D mask data associated with the image.

    window_sizes : List[int]
        A list of window sizes for which the collage method will be applied.

    save_path : str or Path
        The directory path where the resulting features will be saved.

    out_name : str
        The base name used in the saved feature files.

    padding : Union[bool, int], optional (default=False)
        If provided and set to True, padding of 11 is applied. If an integer is provided,
        it determines the amount of padding around the most significant slices in the mask.
        No padding is applied if set to False.

    Returns:
    --------
    None

    Notes:
    ------
    - The resulting features will be saved in two formats: `.pkl` and `.npy`.
    - The names of the saved files will contain the `out_name` and window size.
    """
    if padding:
        if isinstance(padding, bool):
            padding = 11

        most_value_nonzero_slices, _ = significant_slices(mask)

        if most_value_nonzero_slices.size == 0:
            raise ValueError(
                "No significant slices found in mask, mask is either empty or not binary"
            )

        min = most_value_nonzero_slices.min() - padding
        max = most_value_nonzero_slices.max() + padding
        print(f"min: {min}, max: {max}")

        mask = mask[:, :, min:max]
        image = image[:, :, min:max]

    for ws in window_sizes:
        print(f"Processing collage with window size {ws}")
        feats = compute_collage3d(
            image,
            mask,
            haralick_windows=ws,
        )

        print("Final stats", feats)

        if not Path(save_path).exists():
            Path(save_path).mkdir(parents=True, exist_ok=True)

        save_path_pickle = Path(save_path) / f"Feats_Col_{out_name}_ws_{ws}.pkl"
        save_path_npy = Path(save_path) / f"Feats_Col_{out_name}_ws_{ws}.npy"

        with open(save_path_pickle, "wb") as file:
            pickle.dump(feats, file)
        np.save(save_path_npy, feats)



import nibabel as nib
import numpy as np
from scipy.ndimage import zoom

from ..utils import path_in, save_path_file


def resample(input_path, output_path, new_voxel_size, method):
    input_path = path_in(input_path)

    # Load the input NIfTI image
    img = nib.load(input_path)
    data = img.get_fdata()
    affine = img.affine

    header = img.header
    voxel_dims = header.get_zooms()[:3]
    current_voxel_size = voxel_dims

    print("Current voxel size", current_voxel_size)
    current_voxel_size = np.array(current_voxel_size)

    scaling_factors = current_voxel_size / new_voxel_size
    print("Scaling factors", scaling_factors)

    # Compute the new shape after resampling
    print("Input shape", data.shape)
    new_shape = np.ceil(data.shape * scaling_factors).astype(int)
    print("Output shape", new_shape)

    if method == "trilinear":
        # Use Trilinear interpolation to resample the data
        resampled_data = zoom(data, scaling_factors, order=1)
    elif method == "nearest":
        resampled_data = np.zeros(new_shape, dtype=data.dtype)

        # Iterate over each voxel in the resampled image
        for i in range(new_shape[0]):
            for j in range(new_shape[1]):
                for k in range(new_shape[2]):
                    # Compute the corresponding voxel indices in the original image
                    orig_i = int(i / scaling_factors[0])
                    orig_j = int(j / scaling_factors[1])
                    orig_k = int(k / scaling_factors[2])

                    if (
                        0 <= orig_i < data.shape[0]
                        and 0 <= orig_j < data.shape[1]
                        and 0 <= orig_k < data.shape[2]
                    ):
                        resampled_data[i, j, k] = data[orig_i, orig_j, orig_k]

    # Adjust affine to reflect changed voxel spacing
    new_affine = nib.affines.rescale_affine(
        affine, resampled_data.shape, new_voxel_size
    )

    # Create a new NIfTI image with resampled data and the same affine
    resampled_img = nib.Nifti1Image(resampled_data, new_affine)

    # Save the resampled image to the output path
    save_path = save_path_file(output_path, suffix=".nii")

    nib.save(resampled_img, save_path)



import matplotlib.pyplot as plt
import numpy as np
import SimpleITK as sitk


def hu_hist(image_path):
    # Read the image
    image = sitk.ReadImage(image_path)

    # Convert the image to a NumPy array
    image_array = sitk.GetArrayFromImage(image)

    # Create a histogram
    plt.figure(figsize=(10, 6))
    plt.hist(image_array.flatten(), bins=50, color="c")
    plt.title("Histogram of Hounsfield Units (HU)")
    plt.xlabel("Hounsfield Units (HU)")
    plt.ylabel("Frequency")
    plt.grid(True)
    plt.show()



import nibabel as nib
import numpy as np

from .custom_type import TupleNp
from .helper_reader import im2arr


def mask_characteristics(mask_path) -> dict:
    """
    This function takes in a path to a mask and returns a dictionary of characteristics of the mask.
    The characteristics are:
    - shape: the shape of the mask
    - data_type: the data type of the mask
    - values: the unique values in the mask and their counts
    - mask_type: the type of mask (empty, binary, or multi-class)
    :param mask_path: path to the mask
    :return: a dictionary of characteristics of the mask
    """

    # nib_mask = nib.load(mask_path)
    # mask = nib_mask.get_fdata()
    mask = im2arr(mask_path)

    shape = mask.shape
    data_type = mask.dtype

    unique, counts = np.unique(mask, return_counts=True)
    values = dict(zip(unique, counts))

    if len(unique) == 1:
        mask_type = "Empty"
    elif len(unique) == 2:
        mask_type = "Binary"
    else:
        mask_type = "Multi-class"

    mask_characteristics = {
        "shape": shape,
        "data_type": data_type,
        "values": values,
        "mask_type": mask_type,
    }
    return mask_characteristics


def significant_slices(mask) -> TupleNp:
    mask = im2arr(mask)

    if mask_characteristics(mask)["mask_type"] != "Binary":
        print(f"Mask {mask} is not binary.")
        return np.array([]), 0
    else:
        mask_bool = mask.astype(np.bool_)

    z_sum = np.sum(mask_bool, axis=(0, 1))

    most_value_slices = np.argsort(z_sum)[::-1]

    num_nonzero_slices = np.count_nonzero(z_sum)

    most_value_nonzero_slices = most_value_slices[:num_nonzero_slices]

    return most_value_nonzero_slices, num_nonzero_slices


def slice_stats(mask):
    most_value_nonzero_slices, _ = significant_slices(mask)
    min = most_value_nonzero_slices[-1]
    max = most_value_nonzero_slices[0]
    median = most_value_nonzero_slices[len(most_value_nonzero_slices) // 2]
    return {"min": min, "max": max, "median": median}


def significant_slice_idx(mask_path) -> tuple:
    """
    This function takes in a path to a mask and returns the most significant slice index and the number of nonzero slices.
    The most significant slice is the slice with the most nonzero pixels.
    The number of nonzero slices is the number of slices with at least one nonzero pixel.
    :param mask_path: path to the mask
    :return: a tuple of the most significant slice index and the number of nonzero slices
    """
    mask_nb = nib.load(mask_path)
    mask = mask_nb.get_fdata()

    if mask_characteristics(mask_path)["mask_type"] != "Binary":
        print(f"Mask {mask_path} is not binary.")
    else:
        mask = mask.astype(np.bool_)

    most_value_nonzero_slices, num_nonzero_slices = significant_slice_idx_data(mask)

    return most_value_nonzero_slices, num_nonzero_slices


def significant_slice_idx_data(mask_bool) -> tuple:
    """_summary_

    Args:
        mask_bool (_type_): _description_

    Returns:
        tuple: _description_
    """
    z_sum = np.sum(mask_bool, axis=(0, 1))

    most_value_slices = np.argsort(z_sum)[::-1]

    num_nonzero_slices = np.count_nonzero(z_sum)

    most_value_nonzero_slices = most_value_slices[:num_nonzero_slices]

    return most_value_nonzero_slices, num_nonzero_slices


def convert_mask(mask_path, true_value=1):
    mask = nib.load(mask_path)
    mask_data = mask.get_fdata()
    mask_data_bool = np.where(mask_data == true_value, True, False)
    return mask_data_bool


# def mask_path_to_data_ax(mask_path):
#     mask = nib.load(mask_path)
#     mask_data = mask.get_fdata()

#     mask_data_bool = mask_data.astype(np.bool_)
#     mask_data_bool_rot = np.flip(np.rot90(mask_data_bool, axes=(1, 0)), axis=1)
#     # mask_data = np.ma.masked_where(mask_data == False, mask_data)
#     return mask_data_bool_rot


# def mask_path_to_data_sag(mask_path):
#     mask = nib.load(mask_path)
#     mask_data = mask.get_fdata()

#     mask_data_bool = mask_data.astype(np.bool_)
#     mask_data_bool_rot = np.flip(np.rot90(mask_data_bool, axes=(1, 2)), axis=1)
#     # mask_data = np.ma.masked_where(mask_data == False, mask_data)
#     return mask_data_bool_rot


# def image_path_to_data_ax(mask_path):
#     image = nib.load(mask_path)
#     image_data = image.get_fdata()
#     image_data_rot = np.flip(np.rot90(image_data, axes=(1, 0)), axis=1)
#     return image_data_rot


# def image_path_to_data_sag(mask_path):
#     image = nib.load(mask_path)
#     image_data = image.get_fdata()

#     image_data_rot = np.flip(np.rot90(image_data, axes=(1, 2)), axis=1)
#     return image_data_rot


def expand(mask_path, expand_factor=5):
    mask_data = mask_path_to_data_ax(mask_path)

    most_value_nonzero_slices, _ = significant_slice_idx_data(mask_data)

    for d in most_value_nonzero_slices:
        arr = mask_data[:, :, d]
        indices = np.argwhere(arr)
        for m, n in indices:
            blowup_factor = expand_factor
            for i in range(-blowup_factor, blowup_factor + 1):
                for j in range(-blowup_factor, blowup_factor + 1):
                    mask_data[m + i, n + j, d] = True

    return mask_data


def mask_expand(mask_expert_path, mask_paths, expand_factor=5):
    mask_expert_data = expand(mask_expert_path, expand_factor=expand_factor)

    masks_data = []
    for mask_path in mask_paths:
        mask_data = mask_path_to_data_ax(mask_path)
        mask_data = mask_data * mask_expert_data

        masks_data.append(mask_data)

    return mask_expert_data, masks_data


def zero_out(mask_data: np.ndarray, mask_value: int) -> np.ndarray:
    """Zero out all values in the array except for the given value `mask_value` and 0."""

    mask_data = im2arr(mask_data)
    # Create a copy of the input array to avoid modifying the original array
    mask_data_copy = mask_data.copy()

    # Create a mask for the desired values
    mask = (mask_data_copy == mask_value) | (mask_data_copy == 0)

    # Set all values outside of the mask to 0
    mask_data_copy[~mask] = 0
    mask_data_bool = mask_data_copy.astype(bool)

    return mask_data_bool



from pathlib import Path

import pandas as pd

from ..utils import path_in, save_path_file


def match_image_masks(
    images_path: str or Path,
    masks_path: str or Path,
    image_extension: str or None = None,
    mask_extension: str or None = None,
    image_id_func=lambda x: x,
    mask_id_func=lambda x: x,
    save_path: str or Path = Path("./output/match_image_masks.csv"),
):
    df_dict = {"ID": [], "Image": [], "Mask": []}

    images_path = path_in(images_path, env=False)
    masks_path = path_in(masks_path, env=False)

    if mask_extension is None:
        extensions = set()

        for file_path in masks_path.rglob("*"):
            if file_path.is_file():
                file_extension = file_path.suffix.lower()
                extensions.add(file_extension)

        for extension in extensions:
            if extension in [".nii", ".nii.gz", ".dcm", ".npy"]:
                mask_extension = extension
                break

    if image_extension is None:
        extensions = set()

        for file_path in images_path.rglob("*"):
            if file_path.is_file():
                file_extension = file_path.suffix.lower()

                extensions.add(file_extension)

        for extension in extensions:
            if extension in [".nii", ".nii.gz", ".dcm", ".npy"]:
                image_extension = extension
                break

    image_paths = masks_path.glob(rf"**/*{image_extension}")
    mask_paths = masks_path.glob(rf"**/*{mask_extension}")

    images = {}
    for image_path in image_paths:
        id = int(image_id_func(image_path.stem))
        image_path = str(image_path)
        if id in images.keys():
            images[id].append(image_path)
        else:
            images[id] = [image_path]

    masks = {}
    for mask_path in mask_paths:
        id = int(mask_id_func(mask_path.stem))
        mask_path = str(mask_path)
        if id in masks.keys():
            masks[id].append(mask_path)
        else:
            masks[id] = [mask_path]

    # find unique ids
    unique_ids = set(images.keys()).intersection(set(masks.keys()))

    # find all ids
    all_ids = set(images.keys()).union(set(masks.keys()))

    for id in all_ids:
        df_dict["ID"].append(id)

        if id in images.keys():
            df_dict["Image"].append(images[id])
        else:
            df_dict["Image"].append(None)
        if id in masks.keys():
            df_dict["Mask"].append(masks[id])
        else:
            df_dict["Mask"].append(None)

    df = pd.DataFrame(df_dict, columns=df_dict.keys())

    save_path = save_path_file(save_path, suffix=".csv")

    df.to_csv(save_path, index=False)

    print(f"Number of unique ids: {len(unique_ids)}")
    print(f"Number of all ids: {len(all_ids)}")



import logging
import pickle
from pathlib import Path
from typing import List

import numpy as np
from scipy.stats import kurtosis, skew

from .main import Collage

logger = logging.getLogger()

logger.info("Collage feature extraction")

descriptors = [
    "AngularSecondMoment",  # 0
    "Contrast",  # 1
    "Correlation",  # 2
    "SumOfSquareVariance",  # 3
    "SumAverage",  # 4
    "SumVariance",  # 5
    "SumEntropy",  # 6
    "Entropy",  # 7
    "DifferenceVariance",  # 8
    "DifferenceEntropy",  # 9
    "InformationMeasureOfCorrelation1",  # 10
    "InformationMeasureOfCorrelation2",  # 11
    "MaximalCorrelationCoefficient",  # 12
]


def compute_collage2d(
    image: np.ndarray, mask: np.ndarray, haralick_windows: List[int]
) -> np.ndarray:
    feats = {}

    try:
        collage = Collage(
            image,
            mask,
            svd_radius=5,
            verbose_logging=True,
            num_unique_angles=64,
            haralick_window_size=haralick_windows,
        )

        collage_feats = collage.execute()

        print(collage_feats.shape)

        for collage_idx, descriptor in enumerate(descriptors):
            print(f"Processing collage {descriptor}")
            feat = collage_feats[:, :, collage_idx].flatten()
            feat = feat[~np.isnan(feat)]

            feats[f"col_des_{descriptor}"] = [
                feat.mean(),
                feat.std(),
                skew(feat),
                kurtosis(feat),
            ]

    except ValueError as err:
        print(f"VALUE ERROR- {err}")

    except Exception as err:
        print(f"EXCEPTION- {err}")

    return feats


def collage2d(
    image: np.ndarray,
    mask: np.ndarray,
    window_sizes: List[int],
    save_path,
    out_name: str,
):
    """_summary_collage2d
    Compute collage features for a given image and mask.
    :param image: image to compute collage features for
    :type image: np.ndarray
    :param mask: mask to compute collage features for
    :type mask: np.ndarray
    :param window_sizes: window sizes to compute collage features for
    :type window_sizes: List[int]
    :param save_path: path to save collage features
    :type save_path: str
    :param out_name: name of output collage features
    :type out_name: str

    """
    for ws in window_sizes:
        feats = compute_collage2d(
            image,
            mask,
            haralick_windows=ws,
        )

        print("Final stats", feats)

        if not Path(save_path).exists():
            Path(save_path).mkdir(parents=True, exist_ok=True)

        save_path_pickle = Path(save_path) / f"Feats_Col_{out_name}_ws_{ws}.pkl"
        save_path_npy = Path(save_path) / f"Feats_Col_{out_name}_ws_{ws}.npy"

        with open(save_path_pickle, "wb") as file:
            pickle.dump(feats, file)
        np.save(save_path_npy, feats)



# from .mask import generate_mask_schema
from .match_image_mask import match_image_masks
from .resampling import resample
from .metadata import metadata
from .hu import get_hu
from .series2whole import series2whole



from pathlib import Path

import numpy as np
import pydicom

p = Path("/storage/sync/git/test/multi/unzip")

for pp in p.glob("**/*.dcm"):
    dicom_file = pydicom.dcmread(pp)
    pixel_array = dicom_file.pixel_array
    rescale_slope = dicom_file.RescaleSlope
    rescale_intercept = dicom_file.RescaleIntercept

    # print(rescale_slope, rescale_intercept)
    # HU=Pixel Value×Rescale Slope+Rescale Intercept

    hu_image = pixel_array * rescale_slope + rescale_intercept

    save_path = pp.with_suffix(".npz")
    np.savez_compressed(save_path, hu_image=hu_image)

    # exit()



"""
Please read the LICENSE file for details about the LICENSE of this file.
This script is part of the Collage Radiomics project and is governed by its license- under the BSD 3-Clause License.
For more information please visit https://github.com/radxtools/collageradiomics
"""

import logging
import math
from enum import Enum, IntEnum
from itertools import product

import mahotas as mt
import numpy as np
from scipy import linalg
from skimage.feature import graycomatrix
from skimage.util.shape import view_as_windows

logger = logging.getLogger("collageradiomics")
logger.info("Logging set up.")


def _svd_dominant_angles(dx, dy, dz, svd_radius):
    """Calculate a new numpy image containing the dominant angles for each voxel.

    :param dx: 3D numpy array of the pixel gradient in the x directions
    :type dx: numpy.ndarray

    :returns: image array with the dominant angle calculated at each voxel
    :rtype: numpy.ndarray
    """

    is_3D = dx.shape[2] > 1

    # create rolling windows
    svd_diameter = svd_radius * 2 + 1

    # the first (commented out) version would actually take a patch of one slice above and below,
    # but by convention, the third dimension is simply used for the gradient calculation;
    # the actual collection of nearby gradient values to run through the SVD calculation is
    # still done on a given 2D slice
    # window_shape = (svd_diameter, svd_diameter) + ((3 if is_3D else 1),)
    window_shape = (svd_diameter, svd_diameter, 1)
    logger.info(f"Window patch shape for dominant angle calculation = {window_shape}")
    dx_windows = view_as_windows(dx, window_shape)
    dy_windows = view_as_windows(dy, window_shape)
    dz_windows = view_as_windows(dz, window_shape)

    angles_shape = dx_windows.shape[0:3]
    dominant_angles_array = np.zeros(angles_shape + (2 if is_3D else 1,), np.single)

    # loop through each voxel and use SVD to calculate the dominant angle for that rolling window
    # centered on that x,y,z coordinate
    center_x_range = range(angles_shape[1])
    center_y_range = range(angles_shape[0])
    center_z_range = range(angles_shape[2])
    for x, y, z in product(center_x_range, center_y_range, center_z_range):
        dominant_angles_array[y, x, z, :] = _svd_dominant_angle(
            x, y, z, dx_windows, dy_windows, dz_windows
        )

    return dominant_angles_array


def _svd_dominant_angle(x, y, z, dx_windows, dy_windows, dz_windows):
    """Calculates the dominate angle at the coordinate within the windows.

    :param x: x value of coordinate
    :type x: int
    :param y: y value of coordinate
    :type y: int
    :param z: z value of coordinate
    :type z: int
    :param dx_windows: dx windows of x, y shape to run svd upon (shape = rows, cols, slices, row_radius, col_radius, slice_radius)
    :type dx_windows: numpy.ndarray
    :param dy_windows: dy windows of x, y shape to run svd upon
    :type dy_windows: numpy.ndarray
    :param dz_windows: dy windows of x, y shape to run svd upon
    :type dz_windows: numpy.ndarray

    :returns: dominant angle at x, y
    :rtype: float
    """

    # extract the patch of pixel gradient values for this specific voxel
    dx_patch = dx_windows[y, x, z]
    dy_patch = dy_windows[y, x, z]
    dz_patch = dz_windows[y, x, z]

    is_3D = dx_windows.shape[2] > 1

    # flatten all N gradient values in this patch into an Nxd matrix to pass into svd
    window_area = dx_patch.size
    flattened_gradients = np.zeros((window_area, (3 if is_3D else 2)))
    matrix_order = (
        "F"  # fortran-style to be consistent with original matlab implementation
    )
    flattened_gradients[:, 0] = np.reshape(dx_patch, window_area, order=matrix_order)
    flattened_gradients[:, 1] = np.reshape(dy_patch, window_area, order=matrix_order)
    if is_3D:
        flattened_gradients[:, 2] = np.reshape(
            dz_patch, window_area, order=matrix_order
        )

    # calculate svd
    _, _, v = linalg.svd(flattened_gradients)

    # extract results from the first column (in matlab this would be the first row)
    dominant_y = v[0, 0]
    dominant_x = v[1, 0]

    # calculate the dominant angle for this voxel
    dominant_angle = math.atan2(dominant_y, dominant_x)

    if is_3D:
        # also include the secondary angle
        dominant_z = v[2, 0]
        secondary_angle = math.atan2(
            dominant_z, math.sqrt(dominant_x**2 + dominant_y**2)
        )
        return (dominant_angle, secondary_angle)
    else:
        return dominant_angle


class HaralickFeature(IntEnum):
    """Enumeration Helper For Haralick Features

    :param IntEnum: Enumeration Helper For Haralick Features
    :type IntEnum: HaralickFeature
    """

    AngularSecondMoment = 0
    Contrast = 1
    Correlation = 2
    SumOfSquareVariance = 3
    SumAverage = 4
    SumVariance = 5
    SumEntropy = 6
    Entropy = 7
    DifferenceVariance = 8
    DifferenceEntropy = 9
    InformationMeasureOfCorrelation1 = 10
    InformationMeasureOfCorrelation2 = 11
    MaximalCorrelationCoefficient = 12


class DifferenceVarianceInterpretation(Enum):
    """Feature 10 has two interpretations, as the variance of |x-y|
    or as the variance of P(|x-y|).
    See: https://ieeexplore.ieee.org/document/4309314

    :param Enum: Enumeration Helper For Haralick Features
    :type Enum: DifferenceVarianceInterpretation
    """

    XMinusYVariance = 0
    ProbabilityXMinusYVariance = 1


class Collage:
    """This is the main object in the Collage calculation system. Usage: create a Collage object and then call the :py:meth:`execute` function.

    :param image_array: image to run collage upon
    :type image_array: numpy.ndarray
    :param mask_array: mask that correlates with the image
    :type mask_array: numpy.ndarray
    :param svd_radius: radius of svd. Defaults to 5.
    :type svd_radius: int, optional
    :param verbose_logging: This parameter is now ignored. Please use the python logging module.
    :type verbose_logging: bool, optional
    :param cooccurence_angles: list of angles to use in the cooccurence matrix. Defaults to [x*numpy.pi/4 for x in range(8)]
    :type cooccurence_angles: list, optional
    :param difference_variance_interpretation: Feature 10 has two interpretations, as the variance of |x-y| or as the variance of P(|x-y|).].Defaults to DifferenceVarianceInterpretation.XMinusYVariance.
    :type difference_variance_interpretation: DifferenceVarianceInterpretation, optional
    :param haralick_window_size: size of rolling window for texture calculations. Defaults to -1.
    :type haralick_window_size: int, optional
    :param num_unique_angles: number of bins to use for the texture calculation. Defaults to 64.
    :type num_unique_angles: int, optional
    """

    @property
    def img_array(self):
        """
        The original image.

        :getter: Returns the original image array.
        :setter: Sets the original image array.
        :type: np.ndarray
        """
        return self._img_array

    @property
    def mask_array(self):
        """
        Array passed into Collage.

        :getter: Returns the original mask array.
        :setter: Sets the original mask array.
        :type: np.ndarray
        """
        return self._mask_array

    @property
    def is_3D(self):
        """
        Whether we are using 3D collage calculations (True) or 2D (False)
        """
        return self._is_3D

    @property
    def svd_radius(self):
        """
        SVD radius is used to calculate the pixel radius
        for the dominant angle calculation.

        :getter: Returns the SVD radius.
        :setter: Sets the SVD radius.
        :type: int
        """
        return self._svd_radius

    @property
    def verbose_logging(self):
        """
        This parameter is now ignored. Please use the python logging module.

        :getter: Returns True if on.
        :setter: Turns verbose logging off or on.
        :type: bool
        """
        return self._verbose_logging

    @property
    def cooccurence_angles(self):
        """
        Iterable of angles that will be used in the cooccurence matrix.

        :getter: Returns the Iterable of cooccurence angles.
        :setter: Sets the angles to be used in the cooccurence matrix.
        :type: int
        """
        return self._cooccurence_angles

    @property
    def difference_variance_interpretation(self):
        """
        Feature 10 has two interpretations, as the variance of |x-y| or as the variance of P(|x-y|).].
        Defaults to DifferenceVarianceInterpretation.XMinusYVariance.

        :getter: Returns requested variance interpretation.
        :setter: Sets requested variance interpretation.
        :type: DifferenceVarianceInterpretation
        """
        return self._difference_variance_interpretation

    @property
    def haralick_window_size(self):
        """
        Number of pixels around each pixel to calculate a haralick texture.

        :getter: Returns requested number of pixels.
        :setter: Sets requested number of pixels.
        :type: int
        """
        return self._haralick_window_size

    @property
    def num_unique_angles(self):
        """
        Number of bins to use for texture calculations. Defaults to 64.

        :getter: Returns requested number of unique angles to bin into.
        :type: int
        """
        return self._num_unique_angles

    @property
    def collage_output(self):
        """
        Array representing collage upon the mask within the full images.
        If the input was 2D, the output will be height×width×13 where "13" is the number of haralick textures.
        If the input was 3D, the output will be height×width×depth×13x2 where "2" is the primary angle (element 0) or the secondary angle (element 1)

        The output will have numpy.nan values everywhere outside the masked region.

        :getter: Returns array the same shape as the original image with collage in the mask region.
        :type: numpy.ndarray
        """
        return self._collage_output

    @collage_output.setter
    def collage_output(self, value):
        self._collage_output = value

    def get_single_feature_output(self, which_feature):
        """
        Output a single collage output feature.
        If this was a 3D calculation, the output will be of size height×width×depth×2
        where the "2" represents the collage calculation from the primary angle (0) or secondary angle (1).

          :param which_feature: Either an integer from 0 to 12 (inclusive) or a HaralickFeature enum value
          :type which_feature HaralickFeature
        """
        if self.is_3D:
            return self.collage_output[:, :, :, which_feature, :]
        else:
            return self.collage_output[:, :, which_feature]

    def __init__(
        self,
        img_array,
        mask_array,
        svd_radius=5,
        verbose_logging=False,
        cooccurence_angles=[x * np.pi / 4 for x in range(8)],
        difference_variance_interpretation=DifferenceVarianceInterpretation.XMinusYVariance,
        haralick_window_size=-1,
        num_unique_angles=64,
    ):
        """Designated initializer for Collage

        :param image_array: image to run collage upon
        :type image_array: numpy.ndarray
        :param mask_array: mask that correlates with the image
        :type mask_array: numpy.ndarray
        :param svd_radius: radius of svd. Defaults to 5.
        :type svd_radius: int, optional
        :param verbose_logging: This parameter is now ignored. Please use the python logging module.
        :type verbose_logging: bool, optional
        :param cooccurence_angles: list of angles to use in the cooccurence matrix. Defaults to [x * np.pi/4 for x in range(8)]
        :type cooccurence_angles: list, optional
        :param difference_variance_interpretation: Feature 10 has two interpretations, as the variance of |x-y| or as the variance of P(|x-y|).].Defaults to DifferenceVarianceInterpretation.XMinusYVariance.
        :type difference_variance_interpretation: DifferenceVarianceInterpretation, optional
        :param haralick_window_size: size of rolling window for texture calculations. Defaults to -1.
        :type haralick_window_size: int, optional
        :param num_unique_angles: number of bins to use for the texture calculation. Defaults to 64.
        :type num_unique_angles: int, optional
        """

        logger.debug("Collage Module Initialized")

        # error checking
        if haralick_window_size == -1:
            self._haralick_window_size = svd_radius * 2 + 1
        else:
            self._haralick_window_size = haralick_window_size

        if self._haralick_window_size < 1:
            raise Exception("Haralick windows size must be at least 1 pixel.")

        if svd_radius < 1:
            raise Exception("SVD radius must be at least 1 pixel")

        if num_unique_angles < 1:
            raise Exception("num_unique_angles must contain at least 1 bin")

        if img_array.ndim < 2 or img_array.ndim > 3:
            raise Exception("Expected a 2D or 3D image.")

        if mask_array.shape != img_array.shape:
            raise Exception("Mask must be the same shape as image.")

        # Our minimum size for x & y is 50x50
        min_x_y_size = 50

        if img_array.shape[0] < min_x_y_size or img_array.shape[1] < min_x_y_size:
            raise Exception(
                f"Image size ({img_array.shape[0]}x{img_array.shape[1]}) unsupported. Image must be a minimum of a 50x50 for collage to run."
            )

        self._is_3D = img_array.ndim == 3
        logger.debug(f"Running 3D Collage = {self.is_3D}")

        self._img_array = img_array
        if not self.is_3D:
            # in the case of a single 2D slice, give it a third dimension of unit length
            self._img_array = self._img_array.reshape(self._img_array.shape + (1,))

        min_3D_slices = 3
        if (
            self._img_array.shape[0] < self._haralick_window_size
            or self._img_array.shape[1] < self._haralick_window_size
            or (self._is_3D and self._img_array.shape[2] < min_3D_slices)
        ):
            raise Exception(
                f"Image is too small for a window size of {self._haralick_window_size} pixels."
            )

        # threshold mask
        uniqueValues = np.unique(mask_array)
        numberOfValues = len(uniqueValues)
        if numberOfValues > 2:
            logger.info(
                f"Warning: Mask is not binary. Considering all {numberOfValues} nonzero values in the mask as a value of True."
            )
        thresholded_mask_array = mask_array != 0

        # make correct shape
        thresholded_mask_array = thresholded_mask_array.reshape(self.img_array.shape)

        # extract rectangular area of mask
        non_zero_indices = np.argwhere(thresholded_mask_array)
        min_mask_coordinates = non_zero_indices.min(0)
        max_mask_coordinates = non_zero_indices.max(0) + 1
        self.mask_min_x = min_mask_coordinates[1]
        self.mask_min_y = min_mask_coordinates[0]
        self.mask_min_z = min_mask_coordinates[2]
        self.mask_max_x = max_mask_coordinates[1]
        self.mask_max_y = max_mask_coordinates[0]
        self.mask_max_z = max_mask_coordinates[2]

        cropped_mask_array = thresholded_mask_array[
            self.mask_min_y : self.mask_max_y,
            self.mask_min_x : self.mask_max_x,
            self.mask_min_z : self.mask_max_z,
        ]

        # store variables internally
        self._mask_array = cropped_mask_array

        self._svd_radius = svd_radius
        self._verbose_logging = verbose_logging

        self._cooccurence_angles = cooccurence_angles
        self._difference_variance_interpretation = difference_variance_interpretation

        self._num_unique_angles = num_unique_angles

    def _calculate_haralick_feature_values(self, img_array, center_x, center_y):
        """Gets the haralick texture feature values at the x, y, z coordinate.
        , pos[1]
                    :param image_array: image to calculate texture
                    :type image_array: numpy.ndarray
                    :param center_x: x center of coordinate
                    :type center_x: int
                    :param center_y: y center of coordinate
                    :type center_y: int
                    :param window_size: size of window to pull for calculation
                    :type window_size: int
                    :param num_unique_angles: number of bins
                    :type num_unique_angles: int
                    :param haralick_feature: desired haralick feature
                    :type haralick_feature: HaralickFeature

                    :returns: A 13x1 vector of haralick texture at the coordinate.
                    :rtype: numpy.ndarray
        """
        # extract subpart of image (todo: pass in result from view_as_windows)
        window_size = self.haralick_window_size
        min_x = int(max(0, center_x - window_size / 2 - 1))
        min_y = int(max(0, center_y - window_size / 2 - 1))
        max_x = int(min(img_array.shape[1] - 1, center_x + window_size / 2 + 1))
        max_y = int(min(img_array.shape[0] - 1, center_y + window_size / 2 + 1))
        cropped_img_array = img_array[min_y:max_y, min_x:max_x]

        # co-occurence matrix of all 8 directions and sum them
        cooccurence_matrix = graycomatrix(
            cropped_img_array,
            [1],
            self.cooccurence_angles,
            levels=self.num_unique_angles,
        )
        cooccurence_matrix = np.sum(cooccurence_matrix, axis=3)
        cooccurence_matrix = cooccurence_matrix[:, :, 0]

        # extract haralick using mahotas library
        return mt.features.texture.haralick_features(
            [cooccurence_matrix], return_mean=True
        )

    def _calculate_haralick_textures(self, dominant_angles):
        """Gets haralick texture values

        :param dominant_angles_array: An image of the dominant angles at each voxel
        :type dominant_angles_[:,:,:,feature_index]array: numpy.ndarray
        :param desired_haralick_feature: which feature to calculate
        :type desired_haralick_feature: Haralick Feature
        :param num_unique_angles: number of bins
        :type num_unique_angles: int
        :param haralick_window_size: size of window around pixels to calculate haralick value

        :returns: An hxwxdx13 set of haralick texture.
        :rtype: numpy.ndarray
        """

        # rescale from 0 to (num_unique_angles-1)
        num_unique_angles = self.num_unique_angles
        logger.debug(f"Rescaling dominant angles to {num_unique_angles} unique values.")
        dominant_angles_max = dominant_angles.max()
        dominant_angles_min = dominant_angles.min()
        dominant_angles_binned = (
            (dominant_angles - dominant_angles_min)
            / (dominant_angles_max - dominant_angles_min + np.finfo(float).eps)
            * (num_unique_angles - 1)
        )
        dominant_angles_binned = np.round(dominant_angles_binned).astype(int)
        logger.debug(f"Rescaling dominant angles done.")

        # prepare output
        shape = dominant_angles_binned.shape
        haralick_image = np.empty(shape + (13,))
        haralick_image[:] = np.nan

        # the haralick is calculated for each slice separately
        height, width, depth = shape

        logger.debug(
            f"dominant_angles_binned shape is {shape} mask shape is {self.mask_array.shape}"
        )

        # In 3D, we extended the dominant angles by one slice in each direction, so now we need to trim those off.
        for z in range(1, depth - 1) if self.is_3D else range(depth):
            for y, x in product(range(height), range(width)):
                if self.mask_array[y, x, z]:
                    haralick_image[
                        y, x, z, :
                    ] = self._calculate_haralick_feature_values(
                        dominant_angles_binned[:, :, z], x, y
                    )

        return haralick_image

    def execute(self):
        """Begins haralick calculation.

        :returns: An image at original size that only has the masked section filled in with collage calculations.
        :rtype: numpy.ndarray
        """

        svd_radius = self.svd_radius

        # mask location
        mask_min_x = int(self.mask_min_x)
        mask_min_y = int(self.mask_min_y)
        mask_min_z = int(self.mask_min_z)
        mask_max_x = int(self.mask_max_x)
        mask_max_y = int(self.mask_max_y)
        mask_max_z = int(self.mask_max_z)

        mask_width = mask_max_x - mask_min_x
        mask_height = mask_max_y - mask_min_y
        mask_depth = mask_max_z - mask_min_z

        img_array = self.img_array

        # extend the mask outwards a bit (up to the edge of the image) to handle the svd radius
        cropped_min_x = max(mask_min_x - svd_radius, 0)
        cropped_min_y = max(mask_min_y - svd_radius, 0)
        cropped_min_z = max(
            mask_min_z - 1, 0
        )  # for 3D, we just extend 1 slice in both directions
        cropped_max_x = min(mask_max_x + svd_radius, img_array.shape[1])
        cropped_max_y = min(mask_max_y + svd_radius, img_array.shape[0])
        cropped_max_z = min(mask_max_z + 1, img_array.shape[2])

        extended_below = mask_min_z > 0
        extended_above = mask_max_z < img_array.shape[2]

        cropped_image = img_array[
            cropped_min_y:cropped_max_y,
            cropped_min_x:cropped_max_x,
            cropped_min_z:cropped_max_z,
        ]

        logger.debug(f"Image shape = {img_array.shape}")
        logger.debug(f"Mask size = {mask_height}x{mask_width}x{mask_depth}")
        logger.debug(f"Image shape (cropped and padded) = {cropped_image.shape}")

        # ensure the image values range from 0-1
        if cropped_image.max() > 1:
            logger.debug(
                f"Note: Dividing image values by {cropped_image.max()} to convert to 0-1 range"
            )
            cropped_image = cropped_image / cropped_image.max()

        # calculate x,y,z gradients
        logger.debug("Calculating pixel gradients:")
        dx = np.gradient(cropped_image, axis=1)
        dy = np.gradient(cropped_image, axis=0)
        dz = np.gradient(cropped_image, axis=2) if self.is_3D else np.zeros(dx.shape)

        if extended_below:
            dx = dx[:, :, 1:]
            dy = dy[:, :, 1:]
            dz = dz[:, :, 1:]

        if extended_above:
            dx = dx[:, :, :-1]
            dy = dy[:, :, :-1]
            dz = dz[:, :, :-1]

        self.dx = dx
        self.dy = dy
        self.dz = dz
        logger.debug("Calculating pixel gradients done.")

        # calculate dominant angles of each patch
        logger.debug(
            f"Calculating dominant gradient angles using SVD for each image patch of size {svd_radius}x{svd_radius}"
        )
        dominant_angles = _svd_dominant_angles(dx, dy, dz, svd_radius)
        self.dominant_angles = dominant_angles
        angles_shape = dominant_angles.shape
        logger.debug("Calculating dominant gradient angles done.")
        logger.debug(f"Dominant angles shape = {angles_shape}")

        # calculate haralick features of the dominant angles
        logger.debug("Calculating haralick features of angles:")
        haralick_features = np.empty(
            angles_shape[0:3]
            + (
                13,
                2 if self.is_3D else 1,
            )
        )
        for angle_index in range(angles_shape[3]):
            logger.info(f"Calculating features for angle {angle_index}:")
            haralick_features[
                :, :, :, :, angle_index
            ] = self._calculate_haralick_textures(dominant_angles[:, :, :, angle_index])
            logger.info(f"Calculating features for angle {angle_index} done.")
        logger.debug("Calculating haralick features of angles done.")

        # prepare an output full of "NaN's"
        collage_output = np.empty(img_array.shape + haralick_features.shape[3:5])
        collage_output[:] = np.nan

        # if a mask covers the whole image, we'll offset the edges as nans
        if mask_height == img_array.shape[0] and mask_width == img_array.shape[1]:
            y_offset = int((img_array.shape[0] - dominant_angles.shape[0]) / 2)
            mask_min_y += y_offset
            mask_max_y -= y_offset
            x_offset = int((img_array.shape[1] - dominant_angles.shape[1]) / 2)
            mask_min_x += x_offset
            mask_max_x -= x_offset

        # insert the haralick output into the correct spot
        collage_output[
            mask_min_y:mask_max_y, mask_min_x:mask_max_x, mask_min_z:mask_max_z, :, :
        ] = haralick_features

        # remove the singleton third dimension from the output
        if not self.is_3D:
            collage_output = np.squeeze(collage_output, 4)
            collage_output = np.squeeze(collage_output, 2)

        # output
        self.collage_output = collage_output
        logger.debug(f"Output shape = {collage_output.shape}")
        return collage_output



"""
Medical image reader

Extensions:
    - DICOM
    - NIFTI
    - MHA
    
"""

from pathlib import Path
from typing import Tuple, Union

import nibabel as nib
import numpy as np
import pydicom as dicom
import SimpleITK as sitk

from .custom_type import MedicalImage, PathMedicalImage, PathNpType
from .helper_path import path_in


def path2loader(path: PathMedicalImage) -> PathMedicalImage:
    if isinstance(path, MedicalImage):
        return path

    path_obj = path_in(path)
    suffixes = path_obj.suffixes

    try:
        if suffixes[-2:] == [".nii", ".gz"] or suffixes[-1:] == [".nii"]:
            image_and_properties = nib.load(path)
            reader = "nibabel"

        elif suffixes[-1:] == [".dcm"]:
            image_and_properties = dicom.dcmread(path)
            reader = "pydicom"
        else:
            image_and_properties = None
            reader = None

        image_and_properties_itk = sitk.ReadImage(str(path))

    except Exception as e:
        print("Error reading image", e)
        raise e

    return (image_and_properties, reader, image_and_properties_itk)


def im2arr(*paths: PathNpType) -> Union[np.ndarray, Tuple[np.ndarray]]:
    """
    Convert one or more medical images to numpy arrays based on their file extensions.

    Parameters:
        *paths (Union[str, Path]): Paths to the medical image files or already loaded numpy arrays.

    Returns:
        Union[np.ndarray, Tuple[np.ndarray]]: If a single path is provided, returns the image data as a numpy array.
                                             If multiple paths are provided, returns a tuple of numpy arrays.

    Raises:
        TypeError: If the provided file format is unsupported.
    """

    def load_image_as_array(path: Union[str, Path]) -> np.ndarray:
        if isinstance(path, np.ndarray):
            return path

        path_obj = Path(path)
        suffixes = path_obj.suffixes

        if suffixes[-2:] == [".nii", ".gz"] or suffixes[-1:] == [".nii"]:
            return nib.load(path_obj).get_fdata()

        elif suffixes[-1:] == [".dcm"]:
            return dicom.dcmread(path_obj).pixel_array

        elif suffixes[-1:] == [".mha"]:
            sitk_image = sitk.ReadImage(
                str(path_obj)
            )  # SimpleITK python wrapper has no support for pathlib.Path
            return sitk.GetArrayFromImage(sitk_image)
        elif suffixes[-1:] == [".npy"]:
            return np.load(path_obj)
        elif suffixes[-1:] == [".npz"]:
            return np.load(path_obj)["arr_0"]
        else:
            raise TypeError(f"Unsupported file type: {path_obj.suffix}")

    arrays = tuple(load_image_as_array(path) for path in paths)
    if len(arrays) == 1:
        return arrays[0]
    return arrays


def image_preprocess(input, norm: bool) -> np.ndarray:
    """
    Preprocesses the given image by converting it into an array, optionally normalizing it.

    Parameters:
    - input (Union[str, np.ndarray, Any]): The input image to preprocess. This can be a path to the image file,
      an image in numpy array format, or any other format supported by the im2arr function.
    - norm (bool): If set to True, the data will be normalized to fall between 0 and 1.

    Returns:
    - np.ndarray: The preprocessed image in numpy array format.

    Note:
    The function assumes that the im2arr function can convert the provided input into a numpy array.
    """

    data = im2arr(input)
    data = data.astype(np.float64)

    if norm:
        min_value = np.min(data)
        max_value = np.max(data)
        data = (data - min_value) / (max_value - min_value)

    return data


def mask_preprocess(input) -> np.ndarray:
    """
    Preprocesses the given mask by converting it into an array and setting its datatype to int8.

    Parameters:
    - input (Union[str, np.ndarray, Any]): The input mask to preprocess. This can be a path to the mask file,
      a mask in numpy array format, or any other format supported by the im2arr function.

    Returns:
    - np.ndarray: The preprocessed mask in numpy array format.

    Note:
    The function assumes that the im2arr function can convert the provided input into a numpy array.
    """

    data = im2arr(input)
    data = data.astype(np.int8)

    return data


def read_image_mask(
    image: Union[str, Path, np.ndarray],
    mask: Union[str, Path, np.ndarray],
    norm: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Reads the medical image and its corresponding mask, ensuring they both have
    the same shape after preprocessing. If a normalization flag is set for the
    image, it will normalize the image data to [0, 1].

    Parameters:
    - image (Union[str, Path, np.ndarray]): Either a path to the medical image
      file (supporting the formats as defined in image_preprocess function) or
      an already loaded numpy array of the image data.
    - mask (Union[str, Path, np.ndarray]): Similarly, either a path to the mask
      file or an already loaded numpy array of the mask.
    - norm (bool, optional): A flag to decide whether to normalize the image data
      to range [0, 1]. Defaults to False.

    Returns:
    - Tuple[np.ndarray, np.ndarray]: A tuple containing the processed image and
      mask as numpy arrays.

    Raises:
    - ValueError: If the processed image and mask have different shapes.

    Note:
    Ensure that the image_preprocess and mask_preprocess functions are
    properly defined and compatible with the inputs you expect for this function.
    """

    image = image_preprocess(image, norm)
    mask = mask_preprocess(mask)

    if image.shape != mask.shape:
        raise ValueError("Image and mask shape mismatch")

    return image, mask



# from pathlib import Path
# from typing import Any, Dict, Optional, Union

# import nibabel as nib
# import pandas as pd
# from tabulate import tabulate

# from ..utils import path_in, save_path_file


# def metadata(
#     image_path: Union[str, Path], save_csv: Optional[Union[str, Path]] = None
# ) -> Dict[str, Any]:
#     """
#     Extracts and optionally saves metadata from a NIFTI image.

#     Parameters:
#         image_path: Union[str, Path] - The path to the NIFTI image.
#         save_csv: Optional[Union[str, Path]] - The path where to save the metadata as a CSV file.

#     Returns:
#         A dictionary containing the metadata.
#     """
#     image_path = path_in(image_path)

#     # needs refactoring, remove Any
#     image_info: Dict[str, Any] = {}

#     try:
#         image = nib.load(image_path)
#     except nib.filebasedimages.ImageFileError:
#         print(f"Error: {image_path} is not a valid NIFTI image.")
#         return image_info

#     image_data = image.get_fdata()
#     image_header = image.header
#     image_obj = image.dataobj
#     affine = image.affine

#     slope_obj, intercept_obj = image_obj.slope, image_obj.inter

#     is_hu_obj = slope_obj == 1 and intercept_obj <= 0

#     dimensions = image_data.shape
#     voxel_size = image_header.get_zooms()[:3]

#     # Get spatial and temporal units
#     spatial_units, temporal_units = image_header.get_xyzt_units()

#     # Get slope and intercept for Hounsfield Unit (HU) conversion
#     slope, intercept = image_header.get_slope_inter()[:2]

#     # Determine if the image is in HU format
#     is_hu_format = slope == 1 and intercept <= 0

#     # Check if the image is isotropic
#     is_isotropic = all(size == voxel_size[0] for size in voxel_size)

#     # Store the information in the dictionary
#     image_info["dimensions"] = dimensions
#     image_info["voxel_size"] = voxel_size
#     image_info["spatial_units"] = spatial_units
#     image_info["temporal_units"] = temporal_units
#     image_info["slope"] = slope
#     image_info["slope_obj"] = slope_obj
#     image_info["intercept"] = intercept
#     image_info["intercept_obj"] = intercept_obj
#     image_info["is_hu_obj"] = is_hu_obj
#     image_info["is_hu"] = is_hu_format
#     image_info["is_isotropic"] = is_isotropic

#     if save_csv:
#         image_info["name"] = Path(image_path).stem

#         keys = list(image_info.keys())

#         values = list(image_info.values())

#         df = pd.DataFrame([values], columns=keys)
#         csv_path = save_path_file(save_csv, suffix=".csv")

#         df.to_csv(csv_path, index=False)
#     else:
#         print(tabulate(image_info.items(), tablefmt="fancy_grid"))

#     return image_info


from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import nibabel as nib
import numpy as np
import pandas as pd
import pydicom
import SimpleITK as sitk
from tabulate import tabulate

from ..utils import path2loader, path_in, save_path_file

# def metadata_nibabel(image_and_properties):
#     image_info: Dict[str, Any] = {}

#     image_data = image_and_properties.get_fdata()
#     image_header = image_and_properties.header
#     image_obj = image_and_properties.dataobj
#     affine = image_and_properties.affine

#     slope_obj, intercept_obj = image_obj.slope, image_obj.inter

#     is_hu_obj = slope_obj == 1 and intercept_obj <= 0

#     dimensions = image_data.shape
#     voxel_size = image_header.get_zooms()[:3]

#     # Get spatial and temporal units
#     spatial_units, temporal_units = image_header.get_xyzt_units()

#     # Get slope and intercept for Hounsfield Unit (HU) conversion
#     slope, intercept = image_header.get_slope_inter()[:2]

#     # Determine if the image is in HU format
#     is_hu_format = slope == 1 and intercept <= 0

#     # Check if the image is isotropic
#     is_isotropic = all(size == voxel_size[0] for size in voxel_size)

#     # Store the information in the dictionary
#     image_info["dimensions"] = dimensions
#     image_info["voxel_size"] = voxel_size
#     image_info["spatial_units"] = spatial_units
#     image_info["temporal_units"] = temporal_units
#     image_info["slope"] = slope
#     image_info["slope_obj"] = slope_obj
#     image_info["intercept"] = intercept
#     image_info["intercept_obj"] = intercept_obj
#     image_info["is_hu_obj"] = is_hu_obj
#     image_info["is_hu"] = is_hu_format
#     image_info["is_isotropic"] = is_isotropic

#     return image_info


def _metadata_nibabel(image_and_properties: nib.Nifti1Image) -> Dict[str, Any]:
    """
    Extracts metadata from a NIfTI image using Nibabel.

    Parameters:
    image_and_properties (nib.Nifti1Image): A NIfTI image object.

    Returns:
    Dict[str, Any]: A dictionary containing image metadata.
    """
    try:
        image_info: Dict[str, Any] = {}

        image_data = image_and_properties.get_fdata()
        image_header = image_and_properties.header
        image_obj = image_and_properties.dataobj
        affine = image_and_properties.affine

        slope_obj, intercept_obj = image_obj.slope, getattr(image_obj, "inter", 0)

        is_hu_obj = slope_obj == 1 and intercept_obj <= 0

        dimensions = image_data.shape
        voxel_size = image_header.get_zooms()[:3]

        # Get spatial and temporal units
        spatial_units, temporal_units = image_header.get_xyzt_units()

        # Get slope and intercept for Hounsfield Unit (HU) conversion
        slope, intercept = image_header.get_slope_inter()[:2]

        # Determine if the image is in HU format
        is_hu_format = slope == 1 and intercept <= 0

        # Check if the image is isotropic
        is_isotropic = np.allclose(voxel_size, voxel_size[0])

        # Store the information in the dictionary
        image_info = {
            "dimensions": dimensions,
            "voxel_size": voxel_size,
            "spatial_units": spatial_units,
            "temporal_units": temporal_units,
            "slope": slope,
            "slope_obj": slope_obj,
            "intercept": intercept,
            "intercept_obj": intercept_obj,
            "is_hu_obj": is_hu_obj,
            "is_hu": is_hu_format,
            "is_isotropic": is_isotropic,
        }

        return image_info
    except AttributeError as e:
        raise ValueError(f"An attribute error occurred: {e}")


def _metadata_pydicom(image_and_properties) -> Dict[str, Any]:
    """
    Extracts extended metadata from a DICOM file using pydicom.

    Parameters:
    file_path (str): The file path to the DICOM image.

    Returns:
    Dict[str, Any]: A dictionary containing extended image metadata.
    """
    try:
        # Read the DICOM file
        ds = image_and_properties
        

        # Extracting basic metadata
        patient_id = getattr(ds, "PatientID", "N/A")
        patient_name = getattr(ds, "PatientName", "N/A")
        study_id = getattr(ds, "StudyID", "N/A")
        study_date = getattr(ds, "StudyDate", "N/A")
        series_number = getattr(ds, "SeriesNumber", "N/A")
        modality = getattr(ds, "Modality", "N/A")
        manufacturer = getattr(ds, "Manufacturer", "N/A")

        # Extracting pixel spacing, slice thickness, image position and orientation if available
        pixel_spacing = getattr(ds, "PixelSpacing", "N/A")
        slice_thickness = getattr(ds, "SliceThickness", "N/A")
        image_position = getattr(ds, "ImagePositionPatient", "N/A")
        image_orientation = getattr(ds, "ImageOrientationPatient", "N/A")

        # Extracting image dimensions
        dimensions = (
            (int(ds.Rows), int(ds.Columns))
            if "Rows" in ds and "Columns" in ds
            else "N/A"
        )

        # Store the information in the dictionary
        image_info = {
            "patient_id": patient_id,
            "patient_name": patient_name,
            "study_id": study_id,
            "study_date": study_date,
            "series_number": series_number,
            "modality": modality,
            "manufacturer": manufacturer,
            "pixel_spacing": pixel_spacing,
            "slice_thickness": slice_thickness,
            "image_position": image_position,
            "image_orientation": image_orientation,
            "dimensions": dimensions,
        }

        return image_info
    except FileNotFoundError:
        raise FileNotFoundError(f"File not found: {file_path}")
    except Exception as e:
        raise RuntimeError(f"An error occurred while reading the DICOM file: {e}")


def _metadata_sitk(image_and_properties) -> Dict[str, Any]:
    """
    Extracts metadata from an image file using SimpleITK.

    Parameters:
    file_path (str): The file path to the image.

    Returns:
    Dict[str, Any]: A dictionary containing image metadata.
    """
    try:
        # Read the image file
        # image = sitk.ReadImage(file_path)
        image = image_and_properties

        # Extracting metadata
        metadata_keys = image.GetMetaDataKeys()
    
        
        # metadata = {key: image.GetMetaData(key) for key in metadata_keys}
        metadata= {}

        # Adding additional information about the image
        metadata["spacing"] = image.GetSpacing()
        metadata["origin"] = image.GetOrigin()
        metadata["direction"] = image.GetDirection()
        metadata["size"] = image.GetSize()
        metadata[
            "number_of_components_per_pixel"
        ] = image.GetNumberOfComponentsPerPixel()

        return metadata
    except RuntimeError as e:
        raise RuntimeError(f"An error occurred while reading the image file: {e}")


# Example usage:
# metadata = metadata_simpleitk('path_to_image.dcm')
# print(metadata)


def metadata(
    image_paths: Union[List[Union[str, Path]], Union[str, Path]],
    save_csv: Optional[Union[str, Path]] = None,
) -> List[Dict[str, Any]]:
    """
    Extracts and optionally saves metadata from one or more NIFTI images.

    Parameters:
        image_paths: Union[List[Union[str, Path]], Union[str, Path]] - The paths to the NIFTI images.
        save_csv: Optional[Union[str, Path]] - The path where to save the metadata as a CSV file.

    Returns:
        A list of dictionaries, each containing the metadata for one image.
    """

    if not isinstance(image_paths, list):
        image_paths = [image_paths]

    all_image_info = []

    for image_path in image_paths:
        image_path = path_in(image_path)
        # image_info: Dict[str, Any] = {}

        # try:
        #     image = nib.load(image_path)
        # except nib.filebasedimages.ImageFileError:
        #     print(f"Error: {image_path} is not a valid NIFTI image.")
        #     continue

        try:
            image_and_properties, reader, image_and_properties_itk = path2loader(
                image_path
            )

        except Exception as e:
            print("Error reading image", e)

            raise e

        print(reader)

        merged_dict = _metadata_sitk(image_and_properties_itk)

        if reader == "nibabel":
            metadata_nib = _metadata_nibabel(image_and_properties)
            merged_dict = {**merged_dict, **metadata_nib}
            

        elif reader == "pydicom":
            metadata_dicom = _metadata_pydicom(image_and_properties)
            merged_dict = {**merged_dict, **metadata_dicom}
            

        # print(merged_dict)

        # break

        #     image_data = image.get_fdata()
        #     image_header = image.header
        #     image_obj = image.dataobj
        #     affine = image.affine

        #     slope_obj, intercept_obj = image_obj.slope, image_obj.inter

        #     is_hu_obj = slope_obj == 1 and intercept_obj <= 0

        #     dimensions = image_data.shape
        #     voxel_size = image_header.get_zooms()[:3]

        #     # Get spatial and temporal units
        #     spatial_units, temporal_units = image_header.get_xyzt_units()

        #     # Get slope and intercept for Hounsfield Unit (HU) conversion
        #     slope, intercept = image_header.get_slope_inter()[:2]

        #     # Determine if the image is in HU format
        #     is_hu_format = slope == 1 and intercept <= 0

        #     # Check if the image is isotropic
        #     is_isotropic = all(size == voxel_size[0] for size in voxel_size)

        # # Store the information in the dictionary
        # image_info["dimensions"] = dimensions
        # image_info["voxel_size"] = voxel_size
        # image_info["spatial_units"] = spatial_units
        # image_info["temporal_units"] = temporal_units
        # image_info["slope"] = slope
        # image_info["slope_obj"] = slope_obj
        # image_info["intercept"] = intercept
        # image_info["intercept_obj"] = intercept_obj
        # image_info["is_hu_obj"] = is_hu_obj
        # image_info["is_hu"] = is_hu_format
        # image_info["is_isotropic"] = is_isotropic
        # image_info["name"] = image_path.stem

        # # ... (rest of your code for extracting metadata) ...

        all_image_info.append(merged_dict)

    if save_csv:
        csv_path = save_path_file(save_csv, suffix=".csv")
        df = pd.DataFrame(all_image_info)
        df.to_csv(csv_path, index=False)

    else:
        for info in all_image_info:
            print(tabulate(info.items(), tablefmt="fancy_grid"))

    return all_image_info
